import base64
import datetime
import hashlib
import io
import json
import mimetypes
import os
import re
import requests
from pathlib import Path
from typing import Any

from PIL import Image


PROMPT_TEMPLATE_VERSION = 2
PROMPT_MAX_SYSTEM_LENGTH = 4000
PROMPT_MAX_USER_LENGTH = 8000
PROMPT_TEMPERATURE_MIN = 0.0
PROMPT_TEMPERATURE_MAX = 2.0
PROMPT_TAGS_MIN = 1
PROMPT_TAGS_MAX = 20
LLM_BASE_URL_MAX_LENGTH = 512
LLM_MODEL_NAME_MAX_LENGTH = 256

DEFAULT_PROMPT_CONFIG = {
    "system_prompt": (
        "你是我的私人媒体库助手，风格自然、真实、有轻微情绪起伏。"
        "只根据图片可见信息，不要臆测地点、人物身份或具体事件。"
        "输出必须是严格 JSON，不要包含多余文字。"
    ),
    "user_prompt": (
        "这是一张我从社交媒体保存的图片。"
        "请用中文、第一人称、带情绪的一句话给出图片标题。"
        "请给出 1 到 {tags_limit} 个标签。"
        "风格要求：更口语、更生活化，避免模板句；"
        "语气可以轻微变化（偶尔克制、偶尔感慨、偶尔俏皮）。"
        "不确定的信息不要写，避免具体人名/地名/时间。"
        "标签用简短词语，不要带 #，允许情绪词。"
        "输出格式：{\"title\": \"...\", \"tags\": [\"...\", \"...\"]}。"
    ),
    "temperature": 0.6,
    "tags_limit": 5,
    "enabled": True,
}

DEFAULT_LLM_CONFIG = {
    "base_url": "",
    "model_name": "",
}


def get_default_prompt_config() -> dict[str, Any]:
    return dict(DEFAULT_PROMPT_CONFIG)


def get_default_llm_config() -> dict[str, Any]:
    return dict(DEFAULT_LLM_CONFIG)


def merge_prompt_config(base: dict[str, Any], override: dict[str, Any] | None) -> dict[str, Any]:
    merged = dict(base)
    if not override:
        return merged
    for key in ("system_prompt", "user_prompt", "temperature", "tags_limit", "enabled"):
        if key in override:
            merged[key] = override[key]
    return merged


def validate_prompt_config(
    payload: Any,
    *,
    partial: bool = False,
    include_enabled: bool = False,
) -> tuple[dict[str, Any] | None, str | None]:
    if not isinstance(payload, dict):
        return None, "配置格式必须是 JSON 对象。"

    cleaned: dict[str, Any] = {}

    def _read_text(field: str, max_length: int) -> tuple[str | None, str | None]:
        if field not in payload:
            if partial:
                return None, None
            return None, f"缺少字段: {field}"
        value = str(payload.get(field, "")).strip()
        if not value:
            return None, f"{field} 不能为空。"
        if len(value) > max_length:
            return None, f"{field} 不能超过 {max_length} 个字符。"
        return value, None

    system_prompt, error = _read_text("system_prompt", PROMPT_MAX_SYSTEM_LENGTH)
    if error:
        return None, error
    if system_prompt is not None:
        cleaned["system_prompt"] = system_prompt

    user_prompt, error = _read_text("user_prompt", PROMPT_MAX_USER_LENGTH)
    if error:
        return None, error
    if user_prompt is not None:
        cleaned["user_prompt"] = user_prompt

    if "temperature" in payload:
        try:
            temperature = float(payload["temperature"])
        except (TypeError, ValueError):
            return None, "temperature 必须是数字。"
        if not (PROMPT_TEMPERATURE_MIN <= temperature <= PROMPT_TEMPERATURE_MAX):
            return None, f"temperature 必须在 {PROMPT_TEMPERATURE_MIN} 到 {PROMPT_TEMPERATURE_MAX} 之间。"
        cleaned["temperature"] = temperature
    elif not partial:
        cleaned["temperature"] = float(DEFAULT_PROMPT_CONFIG["temperature"])

    if "tags_limit" in payload:
        try:
            tags_limit = int(payload["tags_limit"])
        except (TypeError, ValueError):
            return None, "tags_limit 必须是整数。"
        if not (PROMPT_TAGS_MIN <= tags_limit <= PROMPT_TAGS_MAX):
            return None, f"tags_limit 必须在 {PROMPT_TAGS_MIN} 到 {PROMPT_TAGS_MAX} 之间。"
        cleaned["tags_limit"] = tags_limit
    elif not partial:
        cleaned["tags_limit"] = int(DEFAULT_PROMPT_CONFIG["tags_limit"])

    if include_enabled:
        if "enabled" in payload:
            value = payload["enabled"]
            if not isinstance(value, bool):
                return None, "enabled 必须是布尔值。"
            cleaned["enabled"] = value
        elif not partial:
            cleaned["enabled"] = bool(DEFAULT_PROMPT_CONFIG["enabled"])

    return cleaned, None


def validate_llm_config(
    payload: Any,
    *,
    partial: bool = False,
) -> tuple[dict[str, Any] | None, str | None]:
    if not isinstance(payload, dict):
        return None, "配置格式必须是 JSON 对象。"

    cleaned: dict[str, Any] = {}

    if "base_url" in payload or not partial:
        base_url = str(payload.get("base_url", "")).strip()
        if len(base_url) > LLM_BASE_URL_MAX_LENGTH:
            return None, f"base_url 不能超过 {LLM_BASE_URL_MAX_LENGTH} 个字符。"
        if base_url and not (base_url.startswith("http://") or base_url.startswith("https://")):
            return None, "base_url 必须以 http:// 或 https:// 开头。"
        cleaned["base_url"] = base_url

    if "model_name" in payload or not partial:
        model_name = str(payload.get("model_name", "")).strip()
        if len(model_name) > LLM_MODEL_NAME_MAX_LENGTH:
            return None, f"model_name 不能超过 {LLM_MODEL_NAME_MAX_LENGTH} 个字符。"
        cleaned["model_name"] = model_name

    return cleaned, None


def merge_llm_config(base: dict[str, Any], override: dict[str, Any] | None) -> dict[str, Any]:
    merged = dict(base)
    if not override:
        return merged
    for key in ("base_url", "model_name"):
        value = str(override.get(key, "")).strip() if key in override else ""
        if value:
            merged[key] = value
    return merged


def compute_prompt_hash(prompt_config: dict[str, Any]) -> str:
    stable = {
        "system_prompt": str(prompt_config.get("system_prompt") or ""),
        "user_prompt": str(prompt_config.get("user_prompt") or ""),
        "temperature": float(prompt_config.get("temperature", DEFAULT_PROMPT_CONFIG["temperature"])),
        "tags_limit": int(prompt_config.get("tags_limit", DEFAULT_PROMPT_CONFIG["tags_limit"])),
    }
    encoded = json.dumps(stable, ensure_ascii=False, sort_keys=True, separators=(",", ":")).encode("utf-8")
    return hashlib.sha256(encoded).hexdigest()[:16]


class ImageMetadataStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)

    def _load(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {}
        try:
            with self.store_path.open('r', encoding='utf-8') as f:
                data = json.load(f)
                return data if isinstance(data, dict) else {}
        except Exception:
            return {}

    def get(self, key: str) -> dict[str, Any] | None:
        return self._load().get(key)

    def set(self, key: str, value: dict[str, Any], overwrite: bool = True) -> tuple[dict[str, Any], bool]:
        data = self._load()
        if not overwrite and key in data:
            return data[key], False
        data[key] = value
        self._write(data)
        return value, True

    def _write(self, data: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open('w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)


class PromptConfigStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)

    def _load(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {}
        try:
            with self.store_path.open('r', encoding='utf-8') as f:
                data = json.load(f)
                return data if isinstance(data, dict) else {}
        except Exception:
            return {}

    def get(self) -> dict[str, Any] | None:
        data = self._load()
        if not data:
            return None
        validated, error = validate_prompt_config(data, partial=False, include_enabled=True)
        if error:
            return None
        if isinstance(data.get("updated_at"), str):
            validated["updated_at"] = data["updated_at"]
        return validated

    def set(self, value: dict[str, Any]) -> dict[str, Any]:
        payload = dict(value)
        payload["updated_at"] = datetime.datetime.utcnow().isoformat() + "Z"
        self._write(payload)
        return payload

    def reset(self) -> None:
        try:
            self.store_path.unlink(missing_ok=True)
        except Exception:
            pass

    def _write(self, data: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open('w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)


class LLMConfigStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)

    def _load(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {}
        try:
            with self.store_path.open('r', encoding='utf-8') as f:
                data = json.load(f)
                return data if isinstance(data, dict) else {}
        except Exception:
            return {}

    def get(self) -> dict[str, Any] | None:
        data = self._load()
        if not data:
            return None
        validated, error = validate_llm_config(data, partial=False)
        if error:
            return None
        if isinstance(data.get("updated_at"), str):
            validated["updated_at"] = data["updated_at"]
        return validated

    def set(self, value: dict[str, Any]) -> dict[str, Any]:
        payload = dict(value)
        payload["updated_at"] = datetime.datetime.utcnow().isoformat() + "Z"
        self._write(payload)
        return payload

    def reset(self) -> None:
        try:
            self.store_path.unlink(missing_ok=True)
        except Exception:
            pass

    def _write(self, data: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open('w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)


class CaptionService:
    def __init__(
        self,
        model: str | None = None,
        base_url: str | None = None,
        api_key: str | None = None,
    ):
        self.model = model or os.environ.get('TIKLOCAL_LLM_MODEL')
        self.base_url = base_url or os.environ.get('TIKLOCAL_LLM_BASE_URL')
        self.api_key = api_key or os.environ.get('OPENAI_API_KEY')
        if not self.api_key:
            raise RuntimeError("未配置 OPENAI_API_KEY。")
        if not self.model:
            raise RuntimeError("未配置 TIKLOCAL_LLM_MODEL。")
        if self.base_url and "openrouter.ai" in self.base_url and "/api/v1" not in self.base_url:
            raise RuntimeError("OpenRouter base_url 需要包含 /api/v1，例如 https://openrouter.ai/api/v1")

    def generate(
        self,
        image_path: Path,
        tags_limit: int = 5,
        prompt_config: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        data_url = self._to_data_url(image_path)

        effective_prompt = get_default_prompt_config()
        effective_prompt["tags_limit"] = int(tags_limit)
        effective_prompt = merge_prompt_config(effective_prompt, prompt_config)

        tags_limit = int(effective_prompt["tags_limit"])
        temperature = float(effective_prompt["temperature"])
        system_prompt = str(effective_prompt["system_prompt"])
        user_prompt = self._render_user_prompt(str(effective_prompt["user_prompt"]), tags_limit)

        text = self._request_chat_completion(system_prompt, user_prompt, data_url, temperature)
        if self._looks_like_html(text):
            raise RuntimeError("模型返回了 HTML 页面，请检查 base_url 或 model 是否正确。")

        parsed = self._parse_output(text, tags_limit)

        return {
            "title": parsed.get("title", ""),
            "tags": parsed.get("tags", []),
            "style": "first_person_emotion_zh",
            "model": self.model,
            "provider": "openai",
            "base_url": self.base_url or "",
            "created_at": datetime.datetime.utcnow().isoformat() + "Z",
            "prompt_version": PROMPT_TEMPLATE_VERSION,
            "prompt_hash": compute_prompt_hash(effective_prompt),
        }

    def _render_user_prompt(self, template: str, tags_limit: int) -> str:
        rendered = template.replace("{{tags_limit}}", str(tags_limit))
        return rendered.replace("{tags_limit}", str(tags_limit))

    def _to_data_url(self, image_path: Path, max_size: int = 1536, quality: int = 85) -> str:
        """将图片转换为 base64 data URL，自动压缩以减少 token 消耗。

        Args:
            image_path: 图片文件路径
            max_size: 最长边最大像素，默认 1536px
            quality: JPEG 质量 (1-100)，默认 85

        Returns:
            压缩后的 base64 data URL
        """
        with Image.open(image_path) as img:
            # 转换为 RGB（处理 RGBA、灰度等格式）
            if img.mode in ('RGBA', 'LA', 'P'):
                background = Image.new('RGB', img.size, (255, 255, 255))
                if img.mode == 'P':
                    img = img.convert('RGBA')
                background.paste(img, mask=img.split()[-1] if img.mode in ('RGBA', 'LA') else None)
                img = background
            elif img.mode != 'RGB':
                img = img.convert('RGB')

            # 调整尺寸
            width, height = img.size
            if max(width, height) > max_size:
                ratio = max_size / max(width, height)
                new_size = (int(width * ratio), int(height * ratio))
                img = img.resize(new_size, Image.Resampling.LANCZOS)

            # 压缩为 JPEG
            buffer = io.BytesIO()
            img.save(buffer, format='JPEG', quality=quality, optimize=True)
            encoded = base64.b64encode(buffer.getvalue()).decode("ascii")

        return f"data:image/jpeg;base64,{encoded}"

    def _request_chat_completion(
        self,
        system_prompt: str,
        user_prompt: str,
        data_url: str,
        temperature: float,
    ) -> str:
        base_url = (self.base_url or "https://api.openai.com/v1").rstrip("/")
        url = f"{base_url}/chat/completions"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        payload = {
            "model": self.model,
            "temperature": temperature,
            "messages": [
                {"role": "system", "content": system_prompt},
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": user_prompt},
                        {"type": "image_url", "image_url": {"url": data_url}},
                    ],
                },
            ],
        }
        response = requests.post(url, headers=headers, json=payload, timeout=60)
        text = response.text or ""
        if response.status_code >= 400:
            raise RuntimeError(self._parse_error(text) or f"HTTP {response.status_code}")

        if self._looks_like_html(text):
            return text
        try:
            data = response.json()
        except Exception:
            return text

        if isinstance(data, dict) and data.get("error"):
            raise RuntimeError(self._parse_error(data) or "API error")

        return self._extract_text_from_json(data)

    def _extract_text_from_json(self, data: Any) -> str:
        if not isinstance(data, dict):
            return ""
        choices = data.get("choices") or []
        if choices:
            message = choices[0].get("message") or {}
            content = message.get("content")
            if isinstance(content, str):
                return content
        return ""

    def _parse_error(self, data: Any) -> str:
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except Exception:
                return data.strip()
        if isinstance(data, dict):
            err = data.get("error")
            if isinstance(err, dict):
                return err.get("message") or ""
            if isinstance(err, str):
                return err
        return ""

    def _looks_like_html(self, text: str) -> bool:
        if not text:
            return False
        lowered = text.lstrip().lower()
        head = lowered[:400]
        if lowered.startswith("<!doctype") or lowered.startswith("<html"):
            return True
        return "<html" in head or "<head" in head or "<body" in head

    def _parse_output(self, text: str, tags_limit: int) -> dict[str, Any]:
        data = None
        try:
            data = json.loads(text)
        except Exception:
            match = re.search(r"\{.*\}", text, re.S)
            if match:
                try:
                    data = json.loads(match.group(0))
                except Exception:
                    data = None

        title = ""
        tags: list[str] = []

        if isinstance(data, dict):
            title = str(data.get("title") or data.get("caption") or "").strip()
            tags = data.get("tags") or []
        if not title:
            title = text.strip().splitlines()[0] if text.strip() else ""

        if isinstance(tags, str):
            tags = re.split(r"[，,;/\n]+", tags)
        if isinstance(tags, list):
            tags = [str(t).strip() for t in tags if str(t).strip()]
        else:
            tags = []

        # De-dup and clamp
        seen = set()
        cleaned = []
        for tag in tags:
            if tag in seen:
                continue
            seen.add(tag)
            cleaned.append(tag)
            if len(cleaned) >= tags_limit:
                break

        return {"title": title, "tags": cleaned}



import hashlib
import sys
from pathlib import Path
from tiklocal.services import VIDEO_EXTENSIONS, IMAGE_EXTENSIONS


def compute_file_hash(path: Path, algorithm='sha256', chunk_size=8192) -> str | None:
    """计算文件哈希（支持大文件增量计算）"""
    try:
        hasher = hashlib.new(algorithm)
        with path.open('rb') as f:
            while chunk := f.read(chunk_size):
                hasher.update(chunk)
        return hasher.hexdigest()
    except Exception as e:
        print(f"错误: 无法读取文件 {path}: {e}", file=sys.stderr)
        return None


def _get_file_extensions(file_type: str) -> set[str]:
    """根据类型返回文件扩展名集合"""
    if file_type == 'video':
        return VIDEO_EXTENSIONS
    elif file_type == 'image':
        return IMAGE_EXTENSIONS
    else:  # 'all'
        return VIDEO_EXTENSIONS | IMAGE_EXTENSIONS


def _scan_files(root: Path, extensions: set[str]) -> list[Path]:
    """递归扫描指定扩展名的文件"""
    files = []
    for ext in extensions:
        files.extend(root.glob(f'**/*{ext}'))
        files.extend(root.glob(f'**/*{ext.upper()}'))
    return files


def _print_progress(current: int, total: int, prefix: str = '') -> None:
    """进度条显示"""
    width = 28
    filled = int(width * current / total) if total else width
    bar = '█' * filled + '─' * (width - filled)
    percent = (current / total * 100) if total else 100
    sys.stdout.write(f"\r{prefix}[{bar}] {current}/{total} {percent:5.1f}%")
    sys.stdout.flush()


def find_duplicates(
    media_root: Path,
    file_type: str = 'all',
    algorithm: str = 'sha256',
    show_progress: bool = True
) -> dict[str, list[Path]]:
    """
    查找重复文件
    返回：{hash: [path1, path2, ...]}，只包含有重复的组
    """
    extensions = _get_file_extensions(file_type)
    files = _scan_files(media_root, extensions)
    total = len(files)

    if show_progress:
        print(f'扫描到 {total} 个文件，开始计算哈希值...')

    hash_map = {}  # {hash: [path1, path2, ...]}
    processed = 0

    for file_path in files:
        file_hash = compute_file_hash(file_path, algorithm)
        if file_hash:
            hash_map.setdefault(file_hash, []).append(file_path)

        processed += 1
        if show_progress:
            _print_progress(processed, total, prefix='计算中 ')

    if show_progress:
        print()  # 换行

    # 只返回有重复的组（至少2个文件）
    duplicates = {h: paths for h, paths in hash_map.items() if len(paths) > 1}
    return duplicates


def select_files_to_keep(
    duplicate_groups: dict[str, list[Path]],
    strategy: str = 'oldest'
) -> tuple[list[Path], list[Path]]:
    """
    选择要保留和删除的文件
    返回：(to_keep, to_delete)
    """
    to_keep = []
    to_delete = []

    for file_hash, paths in duplicate_groups.items():
        if strategy == 'oldest':
            # 按 mtime 排序，保留最早的
            sorted_paths = sorted(paths, key=lambda p: p.stat().st_mtime)
        elif strategy == 'newest':
            # 保留最新的
            sorted_paths = sorted(paths, key=lambda p: p.stat().st_mtime, reverse=True)
        elif strategy == 'shortest_path':
            # 保留路径最短的（通常在根目录）
            sorted_paths = sorted(paths, key=lambda p: len(str(p)))
        else:
            sorted_paths = paths

        # 第一个保留，其余删除
        to_keep.append(sorted_paths[0])
        to_delete.extend(sorted_paths[1:])

    return to_keep, to_delete


def format_size(size_bytes: int) -> str:
    """格式化文件大小"""
    for unit in ['B', 'KB', 'MB', 'GB']:
        if size_bytes < 1024:
            return f"{size_bytes:.1f} {unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f} TB"


def delete_files(
    files: list[Path],
    dry_run: bool = True,
    show_progress: bool = True
) -> dict:
    """删除文件"""
    total = len(files)
    deleted = 0
    failed = 0
    total_size = sum(f.stat().st_size for f in files if f.exists())

    if dry_run:
        if show_progress:
            print(f'\n【预演模式】将删除以下 {total} 个文件（共 {format_size(total_size)}）：')
            for f in files:
                size = format_size(f.stat().st_size) if f.exists() else '0 B'
                print(f'  - {f} ({size})')
        return {'deleted': 0, 'failed': 0, 'total': total, 'size_freed': 0}

    if show_progress:
        print(f'\n开始删除 {total} 个文件...')

    processed = 0
    for f in files:
        try:
            if f.exists():
                f.unlink()
                deleted += 1
        except Exception as e:
            failed += 1
            if show_progress:
                print(f"\n警告: 删除失败 {f}: {e}", file=sys.stderr)

        processed += 1
        if show_progress:
            _print_progress(processed, total, prefix='删除中 ')

    if show_progress:
        print()

    return {
        'deleted': deleted,
        'failed': failed,
        'total': total,
        'size_freed': total_size
    }


def run_dedupe(
    media_root: Path,
    file_type: str = 'all',
    algorithm: str = 'sha256',
    keep_strategy: str = 'oldest',
    dry_run: bool = True,
    auto_confirm: bool = False
) -> dict:
    """
    重复文件检测与清理的主流程
    """
    # 1. 查找重复文件
    print(f"开始扫描目录: {media_root}")
    print(f"文件类型: {file_type}  |  哈希算法: {algorithm}  |  保留策略: {keep_strategy}")

    duplicates = find_duplicates(media_root, file_type, algorithm, show_progress=True)

    if not duplicates:
        print("\n✓ 未发现重复文件")
        return {'duplicates': 0, 'deleted': 0}

    # 2. 统计信息
    total_duplicates = sum(len(paths) - 1 for paths in duplicates.values())
    total_groups = len(duplicates)
    print(f"\n发现 {total_groups} 组重复文件，共 {total_duplicates} 个重复副本")

    # 3. 选择要删除的文件
    to_keep, to_delete = select_files_to_keep(duplicates, keep_strategy)

    # 4. 显示详细信息
    print(f"\n重复文件详情：")
    for file_hash, paths in list(duplicates.items())[:5]:  # 只显示前5组
        print(f"\n  哈希: {file_hash[:16]}... ({len(paths)} 个文件)")
        for p in paths:
            status = "✓ 保留" if p in to_keep else "✗ 删除"
            size = format_size(p.stat().st_size) if p.exists() else '0 B'
            print(f"    [{status}] {p} ({size})")

    if len(duplicates) > 5:
        print(f"\n  ... 还有 {len(duplicates) - 5} 组未显示")

    # 5. 删除文件
    if dry_run:
        stats = delete_files(to_delete, dry_run=True, show_progress=True)
        print(f"\n提示: 使用 --execute 参数执行实际删除")
    else:
        if not auto_confirm:
            print(f"\n警告: 即将删除 {len(to_delete)} 个文件!")
            confirm = input("确认继续？(yes/no): ").strip().lower()
            if confirm not in ('yes', 'y'):
                print("操作已取消")
                return {'duplicates': total_duplicates, 'deleted': 0}

        stats = delete_files(to_delete, dry_run=False, show_progress=True)
        print(f"\n✓ 删除完成: {stats['deleted']} 个文件，释放 {format_size(stats['size_freed'])} 空间")
        if stats['failed'] > 0:
            print(f"  失败: {stats['failed']} 个文件", file=sys.stderr)

    return {
        'duplicates': total_duplicates,
        'deleted': stats.get('deleted', 0),
        'failed': stats.get('failed', 0)
    }



import os
import sys
import argparse
from pathlib import Path
from waitress import serve
from tiklocal.app import create_app
from tiklocal.thumbs import generate_thumbnails
from tiklocal.paths import get_data_dir

try:
    import yaml
except ImportError:
    yaml = None


def load_config():
    """从配置文件加载配置"""
    config = {}

    # 尝试读取配置文件
    config_paths = [
        Path.home() / '.config' / 'tiklocal' / 'config.yaml',
        Path.home() / '.tiklocal' / 'config.yaml',
    ]

    for config_path in config_paths:
        if config_path.exists():
            if yaml is None:
                print(f"警告: 找到配置文件 {config_path} 但未安装 PyYAML，跳过配置文件", file=sys.stderr)
                break
            try:
                with open(config_path, 'r', encoding='utf-8') as f:
                    config = yaml.safe_load(f) or {}
                break
            except Exception as e:
                print(f"警告: 读取配置文件 {config_path} 失败: {e}", file=sys.stderr)

    return config


def main():
    # 读取配置文件
    config = load_config()

    # 预处理 argv，支持以下形式：
    # 1) tiklocal                      -> serve
    # 2) tiklocal /path                -> serve /path
    # 3) tiklocal --port 9000          -> serve --port 9000
    # 4) tiklocal thumbs /path         -> thumbs /path
    # 5) tiklocal /path thumbs         -> thumbs /path
    argv = sys.argv[1:]
    if '-h' not in argv and '--help' not in argv:
        if 'thumbs' in argv:
            idx = argv.index('thumbs')
            if idx != 0:
                argv.pop(idx)
                argv.insert(0, 'thumbs')
        elif len(argv) == 0 or argv[0] not in ('serve', 'thumbs', 'dedupe'):
            # 默认回退 serve（空参数或第一个不是已知子命令）
            argv.insert(0, 'serve')

    # 解析命令行参数（支持子命令）
    parser = argparse.ArgumentParser(
        description='TikLocal - 本地媒体服务器',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='''
示例:
  tiklocal                                 # 启动服务（默认）
  tiklocal /path/to/media                  # 指定媒体目录
  tiklocal --port 9000                     # 使用指定端口
  tiklocal serve /path --port 9000         # 显式使用 serve 子命令
  tiklocal thumbs /path --overwrite        # 批量生成缩略图
  tiklocal dedupe /path --dry-run          # 查找重复文件（预演）
  tiklocal dedupe /path --execute          # 删除重复，保留最早文件
        '''
    )

    subparsers = parser.add_subparsers(dest='command')

    # serve 子命令
    serve_parser = subparsers.add_parser('serve', help='启动服务器')
    serve_parser.add_argument('media_root', nargs='?', help='媒体文件根目录路径')
    serve_parser.add_argument('--host', default=None, help='服务器监听地址 (默认: 0.0.0.0)')
    serve_parser.add_argument('--port', type=int, default=None, help='服务器端口 (默认: 8000)')
    serve_parser.add_argument('--dev', action='store_true', help='开发模式（启用热重载和调试）')

    # thumbs 子命令
    thumbs_parser = subparsers.add_parser('thumbs', help='批量生成视频缩略图')
    thumbs_parser.add_argument('media_root', nargs='?', help='媒体文件根目录路径（可省略以使用环境变量/配置文件）')
    thumbs_parser.add_argument('--overwrite', action='store_true', help='存在时覆盖重建')
    thumbs_parser.add_argument('--limit', type=int, default=0, help='最多处理多少个（0 表示全部）')

    # dedupe 子命令
    dedupe_parser = subparsers.add_parser('dedupe', help='检测并清理重复文件')
    dedupe_parser.add_argument('media_root', nargs='?', help='媒体文件根目录路径')
    dedupe_parser.add_argument('--type', choices=['video', 'image', 'all'], default='all',
                              help='文件类型（默认：all）')
    dedupe_parser.add_argument('--algorithm', choices=['md5', 'sha256'], default='sha256',
                              help='哈希算法（默认：sha256，更安全）')
    dedupe_parser.add_argument('--keep', choices=['oldest', 'newest', 'shortest_path'], default='oldest',
                              help='保留策略：oldest=最早文件，newest=最新文件，shortest_path=路径最短（默认：oldest）')
    dedupe_parser.add_argument('--dry-run', action='store_true', default=True,
                              help='预演模式，仅显示将删除的文件（默认开启）')
    dedupe_parser.add_argument('--execute', action='store_true',
                              help='执行实际删除（关闭 dry-run）')
    dedupe_parser.add_argument('--auto-confirm', action='store_true',
                              help='自动确认删除，跳过确认提示（危险）')

    args = parser.parse_args(argv)

    # 判断命令类型（无子命令时视为 serve）
    cmd = args.command or 'serve'

    if cmd == 'thumbs':
        media_root = args.media_root or os.environ.get('MEDIA_ROOT') or config.get('media_root')
        if not media_root:
            parser.error('必须指定媒体目录:\n  - tiklocal thumbs /path/to/media\n  - 或设置环境变量: MEDIA_ROOT=/path/to/media')
        media_path = Path(media_root)
        if not media_path.exists() or not media_path.is_dir():
            print(f"错误: 媒体目录不可用: {media_root}", file=sys.stderr)
            sys.exit(1)
        print(f"数据目录: {get_data_dir()}")
        stats = generate_thumbnails(media_path, overwrite=getattr(args, 'overwrite', False), limit=getattr(args, 'limit', 0), show_progress=True)
        # 完成后退出
        return

    if cmd == 'dedupe':
        from tiklocal.dedupe import run_dedupe

        media_root = args.media_root or os.environ.get('MEDIA_ROOT') or config.get('media_root')
        if not media_root:
            parser.error('必须指定媒体目录:\n  - tiklocal dedupe /path/to/media\n  - 或设置环境变量: MEDIA_ROOT=/path/to/media')

        media_path = Path(media_root)
        if not media_path.exists() or not media_path.is_dir():
            print(f"错误: 媒体目录不可用: {media_root}", file=sys.stderr)
            sys.exit(1)

        # --execute 标志会关闭 dry-run
        dry_run = not getattr(args, 'execute', False)

        stats = run_dedupe(
            media_root=media_path,
            file_type=getattr(args, 'type', 'all'),
            algorithm=getattr(args, 'algorithm', 'sha256'),
            keep_strategy=getattr(args, 'keep', 'oldest'),
            dry_run=dry_run,
            auto_confirm=getattr(args, 'auto_confirm', False)
        )
        return

    # serve 路径
    media_root = args.media_root or os.environ.get('MEDIA_ROOT') or config.get('media_root')
    host = args.host or os.environ.get('TIKLOCAL_HOST') or config.get('host', '0.0.0.0')
    port = args.port or int(os.environ.get('TIKLOCAL_PORT', 0)) or config.get('port', 8000)

    # 验证媒体目录
    if not media_root:
        parser.error('必须指定媒体目录:\n  - 通过命令行参数: tiklocal /path/to/media\n  - 通过环境变量: MEDIA_ROOT=/path/to/media\n  - 通过配置文件: ~/.config/tiklocal/config.yaml')

    media_path = Path(media_root)
    if not media_path.exists():
        print(f"错误: 媒体目录不存在: {media_root}", file=sys.stderr)
        sys.exit(1)

    if not media_path.is_dir():
        print(f"错误: 路径不是目录: {media_root}", file=sys.stderr)
        sys.exit(1)

    # 设置环境变量供 Flask 使用
    os.environ['MEDIA_ROOT'] = str(media_path.absolute())

    # 启动服务器
    print(f"启动 TikLocal 服务器...")
    print(f"媒体目录: {media_path.absolute()}")
    print(f"数据目录: {get_data_dir()}")
    print(f"访问地址: http://{host}:{port}")

    app = create_app()
    if getattr(args, 'dev', False):
        # 开发模式：使用Flask内置服务器
        print("⚠️  开发模式已启用（不要在生产环境使用）")
        app.run(host=host, port=port, debug=True, use_reloader=True)
    else:
        # 生产模式：使用Waitress
        serve(app, host=host, port=port)


if __name__ == '__main__':
    main()



import os
import sys
import json
import hashlib
import subprocess as sp
from pathlib import Path
from tiklocal.paths import get_thumbnails_dir, get_thumbs_map_path

AUDIO_EXTENSIONS = {'.mp3', '.flac', '.aac', '.m4a', '.ogg', '.opus', '.wav'}

class ThumbnailService:
    def __init__(self, media_root: Path):
        self.media_root = media_root
        self.thumb_dir = get_thumbnails_dir()
        self.thumb_map_file = get_thumbs_map_path()
        self.placeholder = (
            b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\x0cIDAT\x08\x99c\xf8\xff\xff?\x00\x05\xfe\x02\xfeA\x93\x8a\x1d\x00\x00\x00\x00IEND\xaeB`\x82"
        )

    def _get_thumb_path(self, rel_path: str) -> Path:
        key = hashlib.sha1(rel_path.encode('utf-8', errors='ignore')).hexdigest() + '.jpg'
        return self.thumb_dir / key

    def get_thumbnail(self, rel_path: str) -> tuple[Path | bytes, str]:
        """Returns (file_path_or_bytes, mimetype)"""
        thumb_path = self._get_thumb_path(rel_path)
        
        # Return existing
        if thumb_path.exists():
            return thumb_path, 'image/jpeg'

        # Generate new
        full_path = self.media_root / rel_path
        if full_path.exists():
            if self._generate(full_path, thumb_path):
                return thumb_path, 'image/jpeg'
        
        return self.placeholder, 'image/png'

    def _generate(self, video_path: Path, output_path: Path, timestamp: float = None) -> bool:
        suffix = video_path.suffix.lower()

        # Audio: extract embedded cover art
        if suffix in AUDIO_EXTENSIONS:
            cmd = ['ffmpeg', '-i', str(video_path), '-an', '-vframes', '1', str(output_path), '-y']
            try:
                sp.run(cmd, stdout=sp.DEVNULL, stderr=sp.DEVNULL, timeout=30)
                if output_path.exists() and output_path.stat().st_size > 0:
                    return True
            except Exception:
                pass
            return False

        candidates = [timestamp] if timestamp is not None else [5.0, 1.0, 0.1]

        for t in candidates:
            cmd = [
                'ffmpeg', '-y',
                '-ss', str(max(0.0, float(t))),
                '-i', str(video_path),
                '-frames:v', '1',
                '-vf', 'scale=-1:360:force_original_aspect_ratio=decrease',
                '-q:v', '3',
                str(output_path)
            ]
            try:
                sp.run(cmd, stdout=sp.DEVNULL, stderr=sp.DEVNULL, timeout=30)
                if output_path.exists() and output_path.stat().st_size > 0:
                    return True
            except Exception:
                continue
        return False



from __future__ import annotations

import datetime
import json
import os
import threading
import uuid
from pathlib import Path
from typing import Any


COLLECTIONS_VERSION = 1
MAX_COLLECTION_NAME_LENGTH = 80
MAX_COLLECTION_DESCRIPTION_LENGTH = 280
MAX_COLLECTION_ITEMS_MUTATION = 200


def _utc_now_iso() -> str:
    return datetime.datetime.utcnow().replace(microsecond=0).isoformat() + "Z"


def _normalize_uri(value: Any) -> str:
    text = str(value or "").strip().replace("\\", "/")
    while text.startswith("./"):
        text = text[2:]
    return text


def _normalize_text(value: Any) -> str:
    return str(value or "").strip()


class CollectionStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)
        self._lock = threading.Lock()

    def _read(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {
                "version": COLLECTIONS_VERSION,
                "updated_at": _utc_now_iso(),
                "collections": [],
            }
        try:
            with self.store_path.open("r", encoding="utf-8") as f:
                payload = json.load(f)
        except Exception:
            return {
                "version": COLLECTIONS_VERSION,
                "updated_at": _utc_now_iso(),
                "collections": [],
            }
        return self._normalize_payload(payload)

    def _normalize_payload(self, payload: Any) -> dict[str, Any]:
        data = payload if isinstance(payload, dict) else {}
        raw_collections = data.get("collections")
        collections: list[dict[str, Any]] = []
        if isinstance(raw_collections, list):
            for item in raw_collections:
                normalized = self._normalize_collection(item)
                if normalized:
                    collections.append(normalized)

        collections.sort(key=lambda item: item.get("updated_at", ""), reverse=True)
        return {
            "version": int(data.get("version") or COLLECTIONS_VERSION),
            "updated_at": _normalize_text(data.get("updated_at")) or _utc_now_iso(),
            "collections": collections,
        }

    def _normalize_collection(self, raw: Any) -> dict[str, Any] | None:
        if not isinstance(raw, dict):
            return None
        collection_id = _normalize_text(raw.get("id"))
        if not collection_id:
            return None

        created_at = _normalize_text(raw.get("created_at")) or _utc_now_iso()
        updated_at = _normalize_text(raw.get("updated_at")) or created_at
        name = _normalize_text(raw.get("name"))[:MAX_COLLECTION_NAME_LENGTH]
        if not name:
            return None
        description = _normalize_text(raw.get("description"))[:MAX_COLLECTION_DESCRIPTION_LENGTH]
        cover_uri = _normalize_uri(raw.get("cover_uri"))

        items_raw = raw.get("items")
        items: list[dict[str, str]] = []
        seen: set[str] = set()
        if isinstance(items_raw, list):
            for item in items_raw:
                if not isinstance(item, dict):
                    continue
                uri = _normalize_uri(item.get("uri"))
                if not uri or uri in seen:
                    continue
                seen.add(uri)
                added_at = _normalize_text(item.get("added_at")) or created_at
                items.append({"uri": uri, "added_at": added_at})

        if cover_uri and all(entry["uri"] != cover_uri for entry in items):
            cover_uri = ""
        if not cover_uri and items:
            cover_uri = items[-1]["uri"]

        return {
            "id": collection_id,
            "name": name,
            "description": description,
            "cover_uri": cover_uri,
            "created_at": created_at,
            "updated_at": updated_at,
            "item_count": len(items),
            "items": items,
        }

    def _write(self, payload: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(payload, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)

    def list(self) -> list[dict[str, Any]]:
        with self._lock:
            payload = self._read()
            return [self._collection_copy(item, include_items=False) for item in payload["collections"]]

    def get(self, collection_id: str) -> dict[str, Any] | None:
        key = _normalize_text(collection_id)
        if not key:
            return None
        with self._lock:
            payload = self._read()
            found = self._find_collection(payload["collections"], key)
            if not found:
                return None
            return self._collection_copy(found, include_items=True)

    def create(self, name: str, description: str = "") -> dict[str, Any]:
        clean_name = _normalize_text(name)[:MAX_COLLECTION_NAME_LENGTH]
        if not clean_name:
            raise ValueError("name 不能为空。")
        clean_desc = _normalize_text(description)[:MAX_COLLECTION_DESCRIPTION_LENGTH]

        now = _utc_now_iso()
        item = {
            "id": "col_" + uuid.uuid4().hex[:12],
            "name": clean_name,
            "description": clean_desc,
            "cover_uri": "",
            "created_at": now,
            "updated_at": now,
            "item_count": 0,
            "items": [],
        }
        with self._lock:
            payload = self._read()
            payload["collections"].insert(0, item)
            payload["updated_at"] = now
            self._write(payload)
        return self._collection_copy(item, include_items=True)

    def update(
        self,
        collection_id: str,
        *,
        name: str | None = None,
        description: str | None = None,
        cover_uri: str | None = None,
    ) -> dict[str, Any] | None:
        key = _normalize_text(collection_id)
        if not key:
            return None

        with self._lock:
            payload = self._read()
            found = self._find_collection(payload["collections"], key)
            if not found:
                return None

            if name is not None:
                clean = _normalize_text(name)[:MAX_COLLECTION_NAME_LENGTH]
                if not clean:
                    raise ValueError("name 不能为空。")
                found["name"] = clean
            if description is not None:
                found["description"] = _normalize_text(description)[:MAX_COLLECTION_DESCRIPTION_LENGTH]
            if cover_uri is not None:
                normalized_cover = _normalize_uri(cover_uri)
                if normalized_cover and all(entry["uri"] != normalized_cover for entry in found["items"]):
                    raise ValueError("cover_uri 不在集合条目中。")
                found["cover_uri"] = normalized_cover

            found["item_count"] = len(found["items"])
            if not found["cover_uri"] and found["items"]:
                found["cover_uri"] = found["items"][-1]["uri"]
            if found["cover_uri"] and all(entry["uri"] != found["cover_uri"] for entry in found["items"]):
                found["cover_uri"] = found["items"][-1]["uri"] if found["items"] else ""

            now = _utc_now_iso()
            found["updated_at"] = now
            payload["updated_at"] = now
            payload["collections"].sort(key=lambda item: item.get("updated_at", ""), reverse=True)
            self._write(payload)
            return self._collection_copy(found, include_items=True)

    def delete(self, collection_id: str) -> bool:
        key = _normalize_text(collection_id)
        if not key:
            return False
        with self._lock:
            payload = self._read()
            collections = payload["collections"]
            next_collections = [item for item in collections if item.get("id") != key]
            if len(next_collections) == len(collections):
                return False
            payload["collections"] = next_collections
            payload["updated_at"] = _utc_now_iso()
            self._write(payload)
            return True

    def add_items(self, collection_id: str, uris: list[str]) -> dict[str, Any] | None:
        key = _normalize_text(collection_id)
        if not key:
            return None
        normalized = self._normalize_mutation_uris(uris)
        if not normalized:
            return self.get(collection_id)

        with self._lock:
            payload = self._read()
            found = self._find_collection(payload["collections"], key)
            if not found:
                return None

            seen = {entry["uri"] for entry in found["items"]}
            now = _utc_now_iso()
            changed = False
            for uri in normalized:
                if uri in seen:
                    continue
                found["items"].append({"uri": uri, "added_at": now})
                seen.add(uri)
                changed = True
            if changed:
                found["item_count"] = len(found["items"])
                if not found["cover_uri"]:
                    found["cover_uri"] = found["items"][-1]["uri"]
                found["updated_at"] = now
                payload["updated_at"] = now
                payload["collections"].sort(key=lambda item: item.get("updated_at", ""), reverse=True)
                self._write(payload)
            return self._collection_copy(found, include_items=True)

    def remove_items(self, collection_id: str, uris: list[str]) -> dict[str, Any] | None:
        key = _normalize_text(collection_id)
        if not key:
            return None
        normalized = set(self._normalize_mutation_uris(uris))
        if not normalized:
            return self.get(collection_id)

        with self._lock:
            payload = self._read()
            found = self._find_collection(payload["collections"], key)
            if not found:
                return None

            next_items = [entry for entry in found["items"] if entry["uri"] not in normalized]
            if len(next_items) == len(found["items"]):
                return self._collection_copy(found, include_items=True)

            found["items"] = next_items
            found["item_count"] = len(next_items)
            if found["cover_uri"] and all(entry["uri"] != found["cover_uri"] for entry in next_items):
                found["cover_uri"] = next_items[-1]["uri"] if next_items else ""
            elif not found["cover_uri"] and next_items:
                found["cover_uri"] = next_items[-1]["uri"]

            now = _utc_now_iso()
            found["updated_at"] = now
            payload["updated_at"] = now
            payload["collections"].sort(key=lambda item: item.get("updated_at", ""), reverse=True)
            self._write(payload)
            return self._collection_copy(found, include_items=True)

    def list_for_media(self, uri: str) -> list[dict[str, Any]]:
        key = _normalize_uri(uri)
        if not key:
            return []
        with self._lock:
            payload = self._read()
            result: list[dict[str, Any]] = []
            for item in payload["collections"]:
                if any(entry["uri"] == key for entry in item["items"]):
                    result.append(self._collection_copy(item, include_items=False))
            return result

    def list_item_uris(self, collection_id: str, *, newest_first: bool = True) -> list[str]:
        data = self.get(collection_id)
        if not data:
            return []
        items = data.get("items") if isinstance(data, dict) else []
        if not isinstance(items, list):
            return []
        uris = [str(entry.get("uri") or "") for entry in items if isinstance(entry, dict)]
        uris = [uri for uri in uris if uri]
        if newest_first:
            uris.reverse()
        return uris

    def _normalize_mutation_uris(self, uris: Any) -> list[str]:
        if not isinstance(uris, list):
            return []
        normalized: list[str] = []
        seen: set[str] = set()
        for item in uris:
            uri = _normalize_uri(item)
            if not uri or uri in seen:
                continue
            seen.add(uri)
            normalized.append(uri)
            if len(normalized) >= MAX_COLLECTION_ITEMS_MUTATION:
                break
        return normalized

    def _find_collection(self, collections: list[dict[str, Any]], collection_id: str) -> dict[str, Any] | None:
        for item in collections:
            if item.get("id") == collection_id:
                return item
        return None

    def _collection_copy(self, item: dict[str, Any], *, include_items: bool) -> dict[str, Any]:
        data = {
            "id": str(item.get("id") or ""),
            "name": str(item.get("name") or ""),
            "description": str(item.get("description") or ""),
            "cover_uri": str(item.get("cover_uri") or ""),
            "created_at": str(item.get("created_at") or ""),
            "updated_at": str(item.get("updated_at") or ""),
            "item_count": int(item.get("item_count") or 0),
        }
        if include_items:
            items = item.get("items")
            normalized_items: list[dict[str, str]] = []
            if isinstance(items, list):
                for entry in items:
                    if not isinstance(entry, dict):
                        continue
                    uri = _normalize_uri(entry.get("uri"))
                    if not uri:
                        continue
                    normalized_items.append({
                        "uri": uri,
                        "added_at": _normalize_text(entry.get("added_at")) or data["created_at"] or _utc_now_iso(),
                    })
            data["items"] = normalized_items
        return data



import datetime
import json
import os
import queue
import re
import shutil
import subprocess as sp
import threading
import uuid
from pathlib import Path
from typing import Any
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse


DOWNLOAD_MAX_URL_LENGTH = 2048
DOWNLOAD_MAX_CONCURRENT_LIMIT = 16
DOWNLOAD_HISTORY_LIMIT = 200
COOKIE_MAX_UPLOAD_BYTES = 1024 * 1024
COOKIE_MATCH_MODE = "filename_contains_domain"
COOKIE_FILE_EXTENSIONS = {".txt", ".cookies"}
DOWNLOAD_ENGINES = {"yt-dlp", "gallery-dl"}
DEFAULT_DOWNLOAD_ENGINE = "yt-dlp"
SOURCE_MAP_VERSION = 1

DEFAULT_DOWNLOAD_CONFIG = {
    "enabled": True,
    "max_concurrent": 2,
    "default_to_root": True,
    "allow_playlist": False,
    "cookie_enabled": True,
    "cookie_dir": "~/.tiklocal/cookies",
    "cookie_match_mode": COOKIE_MATCH_MODE,
    "gallery_archive_enabled": True,
    "gallery_archive_file": "~/.tiklocal/gallery-dl-archive.txt",
}

TERMINAL_JOB_STATUS = {"success", "failed", "canceled"}

_PROGRESS_RE = re.compile(r"\[download\]\s+(?P<percent>\d+(?:\.\d+)?)%")
_ETA_RE = re.compile(r"ETA\s+(?P<eta>[0-9:]+)")
_DESTINATION_PATTERNS = [
    re.compile(r"^\[download\] Destination: (?P<path>.+)$"),
    re.compile(r'^\[Merger\] Merging formats into "(?P<path>.+)"$'),
    re.compile(r'^\[ExtractAudio\] Destination: (?P<path>.+)$'),
]
_TRACKING_QUERY_KEYS = {"fbclid", "gclid", "igshid"}
_OLD_TEMPLATE_ID_RE = re.compile(r"\[(?P<id>[^\]]+)\]")


def _utc_now_iso() -> str:
    return datetime.datetime.utcnow().replace(microsecond=0).isoformat() + "Z"


def _to_int(value: Any) -> int | None:
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def _merge_download_config(base: dict[str, Any], override: dict[str, Any] | None) -> dict[str, Any]:
    merged = dict(base)
    if not override:
        return merged
    for key in (
        "enabled",
        "max_concurrent",
        "default_to_root",
        "allow_playlist",
        "cookie_enabled",
        "cookie_dir",
        "cookie_match_mode",
        "gallery_archive_enabled",
        "gallery_archive_file",
    ):
        if key in override:
            merged[key] = override[key]
    return merged


def validate_download_config(payload: Any, *, partial: bool = False) -> tuple[dict[str, Any] | None, str | None]:
    if not isinstance(payload, dict):
        return None, "配置格式必须是 JSON 对象。"

    cleaned: dict[str, Any] = {}

    defaults = dict(DEFAULT_DOWNLOAD_CONFIG)
    env_cookie_dir = str(os.environ.get("TIKLOCAL_COOKIE_DIR", "")).strip()
    if env_cookie_dir:
        defaults["cookie_dir"] = env_cookie_dir

    def _read_bool(field: str) -> str | None:
        if field not in payload:
            if partial:
                return None
            cleaned[field] = bool(defaults[field])
            return None
        value = payload[field]
        if not isinstance(value, bool):
            return f"{field} 必须是布尔值。"
        cleaned[field] = value
        return None

    for field in ("enabled", "default_to_root", "allow_playlist", "cookie_enabled", "gallery_archive_enabled"):
        error = _read_bool(field)
        if error:
            return None, error

    if "max_concurrent" in payload or not partial:
        value = payload.get("max_concurrent", defaults["max_concurrent"])
        max_concurrent = _to_int(value)
        if max_concurrent is None:
            return None, "max_concurrent 必须是整数。"
        if max_concurrent < 0 or max_concurrent > DOWNLOAD_MAX_CONCURRENT_LIMIT:
            return None, f"max_concurrent 必须在 0 到 {DOWNLOAD_MAX_CONCURRENT_LIMIT} 之间。"
        cleaned["max_concurrent"] = max_concurrent

    if "cookie_dir" in payload or not partial:
        cookie_dir = str(payload.get("cookie_dir", defaults["cookie_dir"])).strip()
        if not cookie_dir:
            return None, "cookie_dir 不能为空。"
        cleaned["cookie_dir"] = cookie_dir

    if "cookie_match_mode" in payload or not partial:
        mode = str(payload.get("cookie_match_mode", defaults["cookie_match_mode"])).strip()
        if mode != COOKIE_MATCH_MODE:
            return None, f"cookie_match_mode 仅支持 {COOKIE_MATCH_MODE}。"
        cleaned["cookie_match_mode"] = mode

    if "gallery_archive_file" in payload or not partial:
        archive_file = str(payload.get("gallery_archive_file", defaults["gallery_archive_file"])).strip()
        if not archive_file:
            return None, "gallery_archive_file 不能为空。"
        cleaned["gallery_archive_file"] = archive_file

    return cleaned, None


def validate_download_url(payload: Any) -> tuple[dict[str, Any] | None, str | None]:
    if not isinstance(payload, dict):
        return None, "请求格式必须是 JSON 对象。"

    url = str(payload.get("url", "")).strip()
    if not url:
        return None, "url 不能为空。"
    if len(url) > DOWNLOAD_MAX_URL_LENGTH:
        return None, f"url 长度不能超过 {DOWNLOAD_MAX_URL_LENGTH} 个字符。"

    parsed = urlparse(url)
    if parsed.scheme not in {"http", "https"} or not parsed.netloc:
        return None, "只支持 http/https 格式的 URL。"

    save_mode = str(payload.get("save_mode", "root")).strip() or "root"
    if save_mode != "root":
        return None, "首版仅支持保存到媒体根目录。"

    engine = str(payload.get("engine", DEFAULT_DOWNLOAD_ENGINE)).strip().lower() or DEFAULT_DOWNLOAD_ENGINE
    if engine not in DOWNLOAD_ENGINES:
        return None, "engine 必须是 yt-dlp 或 gallery-dl。"

    cookie_mode = str(payload.get("cookie_mode", "")).strip().lower()
    cookie_file_raw = payload.get("cookie_file")
    cookie_file = str(cookie_file_raw).strip() if cookie_file_raw is not None else ""

    if cookie_file and not cookie_mode:
        cookie_mode = "manual"
    if not cookie_mode:
        cookie_mode = "auto"

    if cookie_mode not in {"auto", "none", "manual"}:
        return None, "cookie_mode 必须是 auto、none 或 manual。"

    if cookie_mode == "manual":
        if not cookie_file:
            return None, "手动 cookie 模式需要 cookie_file。"
        if not is_safe_cookie_filename(cookie_file):
            return None, "cookie_file 非法，仅允许文件名。"
    else:
        cookie_file = ""

    return {
        "url": url,
        "save_mode": save_mode,
        "engine": engine,
        "cookie_mode": cookie_mode,
        "cookie_file": cookie_file,
    }, None


def is_safe_cookie_filename(name: str) -> bool:
    if not name:
        return False
    if name in {".", ".."}:
        return False
    if "/" in name or "\\" in name:
        return False
    if Path(name).name != name:
        return False
    if Path(name).suffix.lower() not in COOKIE_FILE_EXTENSIONS:
        return False
    return True


def _domain_candidates(host: str) -> list[str]:
    host = host.strip().lower().strip(".")
    if not host:
        return []
    parts = host.split(".")
    if len(parts) <= 1:
        return [host]
    candidates = []
    for i in range(0, len(parts) - 1):
        candidates.append(".".join(parts[i:]))
    return candidates


def _normalize_file_rel(value: str) -> str:
    text = str(value or "").strip().replace("\\", "/")
    while text.startswith("./"):
        text = text[2:]
    return text


def _derive_source_domain(url: str) -> str:
    parsed = urlparse(url)
    return (parsed.hostname or "").strip().lower()


def _strip_tracking_query(url: str) -> str:
    parsed = urlparse(url)
    if parsed.scheme not in {"http", "https"} or not parsed.netloc:
        return url
    pairs = parse_qsl(parsed.query, keep_blank_values=True)
    filtered = []
    for key, value in pairs:
        lowered = key.lower()
        if lowered.startswith("utm_"):
            continue
        if lowered in _TRACKING_QUERY_KEYS:
            continue
        filtered.append((key, value))
    cleaned = parsed._replace(query=urlencode(filtered, doseq=True), fragment="")
    return urlunparse(cleaned)


def _normalize_source_meta(meta: dict[str, Any], *, resolved_by: str | None = None) -> dict[str, Any] | None:
    raw = str(meta.get("source_url_raw") or "").strip()
    display = str(meta.get("source_url_display") or "").strip()
    if not raw and not display:
        return None

    raw = raw or display
    display = display or _strip_tracking_query(raw)
    domain = str(meta.get("source_domain") or "").strip().lower() or _derive_source_domain(raw)
    payload = {
        "source_url_raw": raw,
        "source_url_display": display,
        "source_domain": domain,
        "engine": str(meta.get("engine") or "").strip(),
        "job_id": str(meta.get("job_id") or "").strip(),
        "created_at": str(meta.get("created_at") or _utc_now_iso()).strip(),
    }
    if resolved_by:
        payload["resolved_by"] = resolved_by
    payload["url"] = payload["source_url_display"] or payload["source_url_raw"]
    return payload


class DownloadConfigStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)

    def _load(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {}
        try:
            with self.store_path.open("r", encoding="utf-8") as f:
                data = json.load(f)
                return data if isinstance(data, dict) else {}
        except Exception:
            return {}

    def get(self) -> dict[str, Any]:
        data = self._load()
        defaults = dict(DEFAULT_DOWNLOAD_CONFIG)
        env_cookie_dir = str(os.environ.get("TIKLOCAL_COOKIE_DIR", "")).strip()
        if env_cookie_dir:
            defaults["cookie_dir"] = env_cookie_dir
        validated, _ = validate_download_config(data, partial=True)
        merged = _merge_download_config(defaults, validated)
        if isinstance(data.get("updated_at"), str):
            merged["updated_at"] = data["updated_at"]
        return merged

    def set(self, value: dict[str, Any]) -> dict[str, Any]:
        defaults = self.get()
        defaults.pop("updated_at", None)
        merged = _merge_download_config(defaults, value)
        validated, error = validate_download_config(merged, partial=False)
        if error:
            raise ValueError(error)

        payload = dict(validated)
        payload["updated_at"] = _utc_now_iso()
        self._write(payload)
        return payload

    def _write(self, data: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)


class DownloadHistoryStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)

    def get(self) -> list[dict[str, Any]]:
        if not self.store_path.exists():
            return []
        try:
            with self.store_path.open("r", encoding="utf-8") as f:
                data = json.load(f)
                if isinstance(data, list):
                    return [item for item in data if isinstance(item, dict)]
        except Exception:
            return []
        return []

    def save(self, jobs: list[dict[str, Any]]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(jobs[:DOWNLOAD_HISTORY_LIMIT], f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)


class DownloadSourceStore:
    def __init__(self, store_path: Path):
        self.store_path = store_path
        self.store_path.parent.mkdir(parents=True, exist_ok=True)
        self._lock = threading.Lock()

    def _read(self) -> dict[str, Any]:
        if not self.store_path.exists():
            return {"version": SOURCE_MAP_VERSION, "items": {}, "updated_at": _utc_now_iso()}
        try:
            with self.store_path.open("r", encoding="utf-8") as f:
                data = json.load(f)
                if isinstance(data, dict):
                    return data
        except Exception:
            return {"version": SOURCE_MAP_VERSION, "items": {}, "updated_at": _utc_now_iso()}
        return {"version": SOURCE_MAP_VERSION, "items": {}, "updated_at": _utc_now_iso()}

    def _normalized_payload(self, payload: dict[str, Any] | None = None) -> dict[str, Any]:
        data = payload if isinstance(payload, dict) else self._read()
        items = data.get("items")
        if not isinstance(items, dict):
            items = {}
        return {
            "version": _to_int(data.get("version")) or SOURCE_MAP_VERSION,
            "items": {str(k): v for k, v in items.items() if isinstance(v, dict)},
            "updated_at": str(data.get("updated_at") or _utc_now_iso()),
        }

    def _write(self, payload: dict[str, Any]) -> None:
        tmp_path = self.store_path.with_name(self.store_path.name + ".tmp")
        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(payload, f, ensure_ascii=False, indent=2)
        os.replace(tmp_path, self.store_path)

    def get(self, file_rel: str) -> dict[str, Any] | None:
        key = _normalize_file_rel(file_rel)
        if not key:
            return None
        with self._lock:
            payload = self._normalized_payload()
            value = payload["items"].get(key)
            return dict(value) if isinstance(value, dict) else None

    def get_many(self, file_rels: list[str]) -> dict[str, dict[str, Any] | None]:
        normalized = [_normalize_file_rel(item) for item in file_rels]
        normalized = [item for item in normalized if item]
        result: dict[str, dict[str, Any] | None] = {}
        if not normalized:
            return result
        with self._lock:
            payload = self._normalized_payload()
            items = payload["items"]
            for key in normalized:
                value = items.get(key)
                result[key] = dict(value) if isinstance(value, dict) else None
        return result

    def set_many(self, records: dict[str, dict[str, Any]]) -> int:
        cleaned: dict[str, dict[str, Any]] = {}
        for file_rel, meta in records.items():
            key = _normalize_file_rel(file_rel)
            if not key or not isinstance(meta, dict):
                continue
            cleaned[key] = dict(meta)
        if not cleaned:
            return 0
        with self._lock:
            payload = self._normalized_payload()
            payload["items"].update(cleaned)
            payload["updated_at"] = _utc_now_iso()
            self._write(payload)
        return len(cleaned)

    def delete(self, file_rel: str) -> bool:
        key = _normalize_file_rel(file_rel)
        if not key:
            return False
        with self._lock:
            payload = self._normalized_payload()
            if key not in payload["items"]:
                return False
            payload["items"].pop(key, None)
            payload["updated_at"] = _utc_now_iso()
            self._write(payload)
        return True


class DownloadManager:
    def __init__(
        self,
        media_root: Path,
        config_store: DownloadConfigStore,
        history_store: DownloadHistoryStore,
        source_store: DownloadSourceStore | None = None,
    ):
        self.media_root = media_root.resolve()
        self.config_store = config_store
        self.history_store = history_store
        self.source_store = source_store

        self._lock = threading.Lock()
        self._shutdown = threading.Event()
        self._queue: queue.Queue[str | None] = queue.Queue()

        self._jobs: dict[str, dict[str, Any]] = {}
        self._job_order: list[str] = []
        self._cancel_events: dict[str, threading.Event] = {}
        self._processes: dict[str, sp.Popen[str]] = {}
        self._workers: list[threading.Thread] = []

        self._config = self.config_store.get()
        self._load_history()
        self._ensure_workers()

    def probe_dependencies(self) -> dict[str, Any]:
        yt_dlp_path, yt_dlp_version = self._probe_binary("yt-dlp")
        gallery_dl_path, gallery_dl_version = self._probe_binary("gallery-dl")
        ffmpeg_path, _ = self._probe_binary("ffmpeg")

        return {
            "yt_dlp_available": bool(yt_dlp_path),
            "yt_dlp_path": yt_dlp_path or "",
            "yt_dlp_version": yt_dlp_version,
            "gallery_dl_available": bool(gallery_dl_path),
            "gallery_dl_path": gallery_dl_path or "",
            "gallery_dl_version": gallery_dl_version,
            "ffmpeg_available": bool(ffmpeg_path),
            "ffmpeg_path": ffmpeg_path or "",
        }

    def get_config(self) -> dict[str, Any]:
        with self._lock:
            return dict(self._config)

    def update_config(self, patch: dict[str, Any]) -> dict[str, Any]:
        with self._lock:
            merged = _merge_download_config(self._config, patch)
        validated, error = validate_download_config(merged, partial=False)
        if error:
            raise ValueError(error)

        saved = self.config_store.set(validated)
        with self._lock:
            self._config = _merge_download_config(DEFAULT_DOWNLOAD_CONFIG, saved)
            self._config["updated_at"] = saved.get("updated_at")
            self._ensure_workers_locked()
            return dict(self._config)

    def list_cookie_files(self) -> dict[str, Any]:
        with self._lock:
            cookie_enabled = bool(self._config.get("cookie_enabled", True))
            cookie_dir_text = str(self._config.get("cookie_dir", DEFAULT_DOWNLOAD_CONFIG["cookie_dir"])).strip()
        cookie_dir = self._expand_cookie_dir(cookie_dir_text)

        files: list[str] = []
        if cookie_enabled and cookie_dir.exists() and cookie_dir.is_dir():
            for entry in cookie_dir.iterdir():
                if not entry.is_file():
                    continue
                if entry.suffix.lower() not in COOKIE_FILE_EXTENSIONS:
                    continue
                files.append(entry.name)
        files.sort(key=lambda name: name.lower())
        return {
            "cookie_enabled": cookie_enabled,
            "cookie_dir": str(cookie_dir),
            "files": files,
        }

    def enqueue(
        self,
        url: str,
        *,
        save_mode: str = "root",
        engine: str = DEFAULT_DOWNLOAD_ENGINE,
        cookie_mode: str = "auto",
        cookie_file: str = "",
        retry_of: str = "",
        output_token: str = "",
    ) -> dict[str, Any]:
        engine = (engine or DEFAULT_DOWNLOAD_ENGINE).strip().lower()
        if engine not in DOWNLOAD_ENGINES:
            raise RuntimeError("不支持的下载引擎。")

        chosen_file, chosen_mode, cookie_error = self._resolve_cookie_choice(
            url=url,
            cookie_mode=cookie_mode,
            cookie_file=cookie_file,
        )
        if cookie_error:
            raise RuntimeError(cookie_error)

        with self._lock:
            if not self._config.get("enabled", True):
                raise RuntimeError("下载功能已禁用。")

            job_id = uuid.uuid4().hex[:12]
            now = _utc_now_iso()
            job = {
                "id": job_id,
                "url": url,
                "save_mode": save_mode,
                "engine": engine,
                "engine_version": self._probe_binary(engine)[1],
                "status": "queued",
                "progress_percent": None,
                "eta_sec": None,
                "created_at": now,
                "started_at": None,
                "finished_at": None,
                "output_path_rel": "",
                "output_files_rel": [],
                "file_count": 0,
                "error_message": "",
                "requested_format": "mp4_preferred",
                "cancel_requested": False,
                "cookie_file": chosen_file,
                "cookie_match_mode": chosen_mode,
                "retry_of": retry_of.strip(),
                "output_token": output_token.strip() or now.replace(":", "").replace("-", "").replace("Z", ""),
            }
            self._jobs[job_id] = job
            self._job_order.insert(0, job_id)
            self._cancel_events[job_id] = threading.Event()
            self._persist_locked()
            max_concurrent = int(self._config.get("max_concurrent", 2))

        if max_concurrent == 0:
            threading.Thread(target=self._run_job, args=(job_id,), daemon=True).start()
        else:
            self._ensure_workers()
            self._queue.put(job_id)

        return self.get_job(job_id) or job

    def upload_cookie_file(self, filename: str, content: bytes, *, replace: bool = False) -> dict[str, Any]:
        safe_name = str(filename or "").strip()
        if not is_safe_cookie_filename(safe_name):
            raise ValueError("文件名非法，仅支持 .txt/.cookies。")
        if not isinstance(content, (bytes, bytearray)):
            raise ValueError("文件内容格式错误。")
        if len(content) == 0:
            raise ValueError("文件内容不能为空。")
        if len(content) > COOKIE_MAX_UPLOAD_BYTES:
            raise ValueError(f"文件不能超过 {COOKIE_MAX_UPLOAD_BYTES // 1024} KB。")

        cookie_dir = self._cookie_dir_path()
        target = (cookie_dir / safe_name).resolve()
        try:
            target.relative_to(cookie_dir.resolve())
        except ValueError as exc:
            raise ValueError("文件路径非法。") from exc

        if target.exists() and not replace:
            raise ValueError(f"文件已存在: {safe_name}")

        tmp_path = target.with_suffix(target.suffix + ".tmp")
        with tmp_path.open("wb") as f:
            f.write(bytes(content))
        os.replace(tmp_path, target)
        try:
            os.chmod(target, 0o600)
        except Exception:
            pass

        return {
            "filename": safe_name,
            "size": target.stat().st_size if target.exists() else len(content),
            "cookie_dir": str(cookie_dir),
        }

    def delete_job(self, job_id: str) -> tuple[bool, str | None]:
        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return False, "Job not found"
            if job.get("status") not in TERMINAL_JOB_STATUS:
                return False, "运行中任务不可删除"

            self._jobs.pop(job_id, None)
            self._cancel_events.pop(job_id, None)
            self._processes.pop(job_id, None)
            self._job_order = [jid for jid in self._job_order if jid != job_id]
            self._persist_locked()
        return True, None

    def clear_history(self) -> int:
        deleted = 0
        with self._lock:
            keep_ids: list[str] = []
            for job_id in self._job_order:
                job = self._jobs.get(job_id)
                if not job:
                    continue
                if job.get("status") in TERMINAL_JOB_STATUS:
                    deleted += 1
                    self._jobs.pop(job_id, None)
                    self._cancel_events.pop(job_id, None)
                    self._processes.pop(job_id, None)
                else:
                    keep_ids.append(job_id)
            self._job_order = keep_ids
            self._persist_locked()
        return deleted

    def retry_job(self, job_id: str) -> tuple[dict[str, Any] | None, str | None]:
        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return None, "Job not found"
            if job.get("status") not in {"failed", "canceled"}:
                return None, "仅失败或已取消任务支持重试"
            url = str(job.get("url") or "")
            save_mode = str(job.get("save_mode") or "root")
            engine = str(job.get("engine") or DEFAULT_DOWNLOAD_ENGINE)
            cookie_file = str(job.get("cookie_file") or "")
            cookie_match_mode = str(job.get("cookie_match_mode") or "none")
            output_token = str(job.get("output_token") or "")

        cookie_mode = "none"
        if cookie_match_mode in {"auto", "manual"}:
            cookie_mode = cookie_match_mode

        try:
            new_job = self.enqueue(
                url,
                save_mode=save_mode,
                engine=engine,
                cookie_mode=cookie_mode,
                cookie_file=cookie_file,
                retry_of=job_id,
                output_token=output_token,
            )
            return new_job, None
        except RuntimeError as exc:
            return None, str(exc)

    def list_jobs(self, *, limit: int = 50) -> list[dict[str, Any]]:
        limit = max(1, min(int(limit or 50), DOWNLOAD_HISTORY_LIMIT))
        with self._lock:
            job_ids = self._job_order[:limit]
            return [dict(self._jobs[job_id]) for job_id in job_ids if job_id in self._jobs]

    def get_job(self, job_id: str) -> dict[str, Any] | None:
        with self._lock:
            job = self._jobs.get(job_id)
            return dict(job) if job else None

    def resolve_source_for_file(self, file_rel: str) -> dict[str, Any] | None:
        key = _normalize_file_rel(file_rel)
        if not key:
            return None

        mapped = self._resolve_source_from_map(key)
        if mapped:
            return mapped

        info_meta = self._resolve_source_from_info_json(key)
        if info_meta:
            return info_meta

        return self._resolve_source_from_filename(key)

    def resolve_sources_for_files(self, file_rels: list[str]) -> dict[str, dict[str, Any] | None]:
        normalized = [_normalize_file_rel(item) for item in file_rels]
        normalized = [item for item in normalized if item]
        normalized = list(dict.fromkeys(normalized))
        result: dict[str, dict[str, Any] | None] = {}
        for file_rel in normalized:
            result[file_rel] = self.resolve_source_for_file(file_rel)
        return result

    def delete_source_for_file(self, file_rel: str) -> bool:
        if not self.source_store:
            return False
        return self.source_store.delete(file_rel)

    def cancel(self, job_id: str) -> dict[str, Any] | None:
        process: sp.Popen[str] | None = None
        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return None

            if job["status"] in TERMINAL_JOB_STATUS:
                return dict(job)

            cancel_event = self._cancel_events.get(job_id)
            if cancel_event:
                cancel_event.set()

            job["cancel_requested"] = True
            if job["status"] == "queued":
                self._mark_canceled_locked(job)
                return dict(job)

            process = self._processes.get(job_id)

        if process:
            self._terminate_process(process)

        return self.get_job(job_id)

    def _load_history(self) -> None:
        history = self.history_store.get()
        now = _utc_now_iso()
        with self._lock:
            for item in history[:DOWNLOAD_HISTORY_LIMIT]:
                job_id = str(item.get("id") or "").strip()
                if not job_id:
                    continue

                status = str(item.get("status") or "failed")
                if status in {"queued", "running"}:
                    status = "failed"
                    item["error_message"] = "任务因服务重启中断。"
                    item["finished_at"] = now
                raw_output_files = item.get("output_files_rel")
                output_files_rel = raw_output_files if isinstance(raw_output_files, list) else []

                job = {
                    "id": job_id,
                    "url": str(item.get("url") or ""),
                    "save_mode": str(item.get("save_mode") or "root"),
                    "engine": str(item.get("engine") or DEFAULT_DOWNLOAD_ENGINE),
                    "engine_version": str(item.get("engine_version") or ""),
                    "status": status,
                    "progress_percent": item.get("progress_percent"),
                    "eta_sec": item.get("eta_sec"),
                    "created_at": str(item.get("created_at") or now),
                    "started_at": item.get("started_at"),
                    "finished_at": item.get("finished_at"),
                    "output_path_rel": str(item.get("output_path_rel") or ""),
                    "output_files_rel": [str(v) for v in output_files_rel if str(v).strip()],
                    "file_count": _to_int(item.get("file_count")) or 0,
                    "error_message": str(item.get("error_message") or ""),
                    "requested_format": str(item.get("requested_format") or "mp4_preferred"),
                    "cancel_requested": False,
                    "cookie_file": str(item.get("cookie_file") or ""),
                    "cookie_match_mode": str(item.get("cookie_match_mode") or "none"),
                    "retry_of": str(item.get("retry_of") or ""),
                    "output_token": str(
                        item.get("output_token")
                        or str(item.get("created_at") or now).replace(":", "").replace("-", "").replace("Z", "")
                    ),
                }
                if not job["file_count"] and job["output_files_rel"]:
                    job["file_count"] = len(job["output_files_rel"])
                if not job["output_files_rel"] and job["output_path_rel"]:
                    job["output_files_rel"] = [job["output_path_rel"]]
                    job["file_count"] = max(1, job["file_count"])
                self._jobs[job_id] = job
                self._job_order.append(job_id)
                self._cancel_events[job_id] = threading.Event()

    def _ensure_workers(self) -> None:
        with self._lock:
            self._ensure_workers_locked()

    def _ensure_workers_locked(self) -> None:
        max_concurrent = int(self._config.get("max_concurrent", DEFAULT_DOWNLOAD_CONFIG["max_concurrent"]))
        if max_concurrent <= 0:
            return

        self._workers = [worker for worker in self._workers if worker.is_alive()]
        missing = max_concurrent - len(self._workers)
        for _ in range(max(0, missing)):
            worker = threading.Thread(target=self._worker_loop, daemon=True)
            worker.start()
            self._workers.append(worker)

    def _worker_loop(self) -> None:
        while not self._shutdown.is_set():
            try:
                job_id = self._queue.get(timeout=0.5)
            except queue.Empty:
                continue

            if job_id is None:
                self._queue.task_done()
                break

            self._run_job(job_id)
            self._queue.task_done()

    def _run_job(self, job_id: str) -> None:
        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return

            if job["status"] != "queued":
                return

            cancel_event = self._cancel_events.get(job_id)
            if cancel_event and cancel_event.is_set():
                self._mark_canceled_locked(job)
                return

            job["status"] = "running"
            job["started_at"] = _utc_now_iso()
            job["progress_percent"] = 0.0
            self._persist_locked()

        try:
            result = self._execute_download(job_id)
            return_code, error_message, output_rel_list = self._normalize_execute_result(result)
            source_context: dict[str, Any] | None = None

            with self._lock:
                job = self._jobs.get(job_id)
                if not job:
                    return

                cancel_event = self._cancel_events.get(job_id)
                if cancel_event and cancel_event.is_set():
                    self._mark_canceled_locked(job)
                    return

                if return_code == 0:
                    job["status"] = "success"
                    job["progress_percent"] = 100.0
                    job["eta_sec"] = 0
                    job["output_files_rel"] = output_rel_list
                    job["file_count"] = len(output_rel_list)
                    job["output_path_rel"] = output_rel_list[0] if output_rel_list else ""
                    job["error_message"] = ""
                    source_context = {
                        "url": str(job.get("url") or ""),
                        "engine": str(job.get("engine") or ""),
                        "job_id": str(job.get("id") or ""),
                        "created_at": str(job.get("created_at") or _utc_now_iso()),
                        "files": list(output_rel_list),
                    }
                else:
                    job["status"] = "failed"
                    job["error_message"] = error_message or "下载失败，请检查 URL 与网络环境。"

                job["finished_at"] = _utc_now_iso()
                self._persist_locked()

            if source_context:
                self._record_job_sources_on_success(source_context)
        except FileNotFoundError as exc:
            with self._lock:
                job = self._jobs.get(job_id)
                if not job:
                    return
                engine = str(job.get("engine") or DEFAULT_DOWNLOAD_ENGINE)
                job["status"] = "failed"
                job["finished_at"] = _utc_now_iso()
                missing_text = str(exc or "").lower()
                if "gallery-dl" in missing_text or engine == "gallery-dl":
                    job["error_message"] = "未检测到 gallery-dl，请先安装后再使用该引擎。"
                else:
                    job["error_message"] = "未检测到 yt-dlp，请先安装后再使用下载功能。"
                self._persist_locked()
        except Exception as exc:  # pragma: no cover - defensive branch
            with self._lock:
                job = self._jobs.get(job_id)
                if not job:
                    return
                job["status"] = "failed"
                job["finished_at"] = _utc_now_iso()
                job["error_message"] = str(exc)
                self._persist_locked()
        finally:
            with self._lock:
                self._processes.pop(job_id, None)

    def _execute_download(self, job_id: str) -> tuple[int, str, list[str]] | tuple[int, str, str]:
        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return 1, "任务不存在。", ""
            engine = str(job.get("engine") or DEFAULT_DOWNLOAD_ENGINE)

        if engine == "gallery-dl":
            return self._execute_download_gallery_dl(job_id)
        return self._execute_download_yt_dlp(job_id)

    def _execute_download_yt_dlp(self, job_id: str) -> tuple[int, str, list[str]]:
        yt_dlp_bin = shutil.which("yt-dlp")
        if not yt_dlp_bin:
            raise FileNotFoundError("yt-dlp not found")

        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return 1, "任务不存在。", []
            url = job["url"]
            allow_playlist = bool(self._config.get("allow_playlist", False))
            cookie_file = str(job.get("cookie_file") or "")
            cookie_match_mode = str(job.get("cookie_match_mode") or "none")

        output_template = str(
            self.media_root
            / "%(extractor_key|na)s__%(uploader_id|na).24B__%(display_id|na)s__%(id|na)s__%(upload_date|na)s__%(autonumber)02d.%(ext)s"
        )

        cmd = [
            yt_dlp_bin,
            "--newline",
            "--restrict-filenames",
            "--merge-output-format",
            "mp4",
            "--write-info-json",
            "--continue",
            "--retries",
            "10",
            "--fragment-retries",
            "10",
            "--file-access-retries",
            "5",
            "--socket-timeout",
            "30",
            "-o",
            output_template,
        ]
        if not allow_playlist:
            cmd.append("--no-playlist")

        if cookie_file and cookie_match_mode in {"auto", "manual"}:
            cookie_path, error = self._resolve_cookie_path(cookie_file)
            if error:
                return 1, error, []
            cmd.extend(["--cookies", str(cookie_path)])

        cmd.append(url)

        process = sp.Popen(
            cmd,
            stdout=sp.PIPE,
            stderr=sp.STDOUT,
            text=True,
            bufsize=1,
        )

        with self._lock:
            self._processes[job_id] = process

        cancel_event = self._cancel_events.get(job_id)
        output_path_abs = ""
        error_message = ""

        stream = process.stdout
        if stream is not None:
            for raw_line in stream:
                line = raw_line.strip()
                if not line:
                    continue

                progress = self._parse_progress(line)
                if progress:
                    self._update_progress(job_id, progress)

                found_path = self._parse_output_path(line)
                if found_path:
                    output_path_abs = found_path

                extracted_error = self._extract_error_line(line)
                if extracted_error:
                    error_message = extracted_error

                if cancel_event and cancel_event.is_set():
                    self._terminate_process(process)

        return_code = process.wait()
        if not error_message and return_code != 0:
            error_message = f"yt-dlp exited with code {return_code}."

        output_rel = self._to_media_relative(output_path_abs)
        return return_code, error_message, [output_rel] if output_rel else []

    def _execute_download_gallery_dl(self, job_id: str) -> tuple[int, str, list[str]]:
        gallery_dl_bin = shutil.which("gallery-dl")
        if not gallery_dl_bin:
            raise FileNotFoundError("gallery-dl not found")

        with self._lock:
            job = self._jobs.get(job_id)
            if not job:
                return 1, "任务不存在。", []
            url = str(job.get("url") or "")
            cookie_file = str(job.get("cookie_file") or "")
            cookie_match_mode = str(job.get("cookie_match_mode") or "none")
            archive_enabled = bool(self._config.get("gallery_archive_enabled", True))
            archive_file_text = str(self._config.get("gallery_archive_file", DEFAULT_DOWNLOAD_CONFIG["gallery_archive_file"])).strip()

        temp_dir = (self.media_root / ".tiklocal-download-tmp" / job_id).resolve()
        temp_dir.mkdir(parents=True, exist_ok=True)
        log_file = temp_dir / "gallery-dl.log"

        cmd = [
            gallery_dl_bin,
            "--no-colors",
            "--directory",
            str(temp_dir),
            "--retries",
            "10",
            "--http-timeout",
            "30",
            "--sleep-429",
            "8",
        ]

        if archive_enabled:
            archive_path = self._expand_user_path(archive_file_text)
            archive_path.parent.mkdir(parents=True, exist_ok=True)
            cmd.extend(["--download-archive", str(archive_path)])

        if cookie_file and cookie_match_mode in {"auto", "manual"}:
            cookie_path, error = self._resolve_cookie_path(cookie_file)
            if error:
                shutil.rmtree(temp_dir, ignore_errors=True)
                return 1, error, []
            cmd.extend(["--cookies", str(cookie_path)])

        cmd.extend(["--write-log", str(log_file), url])

        process = sp.Popen(
            cmd,
            stdout=sp.PIPE,
            stderr=sp.STDOUT,
            text=True,
            bufsize=1,
        )

        with self._lock:
            self._processes[job_id] = process

        cancel_event = self._cancel_events.get(job_id)
        error_message = ""

        stream = process.stdout
        if stream is not None:
            for raw_line in stream:
                line = raw_line.strip()
                if not line:
                    continue

                extracted_error = self._extract_error_line(line)
                if extracted_error:
                    error_message = extracted_error

                if cancel_event and cancel_event.is_set():
                    self._terminate_process(process)

        return_code = process.wait()

        try:
            if return_code != 0:
                if not error_message:
                    error_message = f"gallery-dl exited with code {return_code}."
                return return_code, error_message, []

            outputs = self._collect_gallery_outputs(temp_dir, excluded={log_file.resolve()})
            moved_outputs = [self._move_file_to_media_root(path) for path in outputs]
            moved_outputs = [path for path in moved_outputs if path]
            if not moved_outputs:
                return 1, "gallery-dl 未下载到可用文件。", []
            return 0, "", moved_outputs
        finally:
            shutil.rmtree(temp_dir, ignore_errors=True)

    def _update_progress(self, job_id: str, progress: dict[str, Any]) -> None:
        with self._lock:
            job = self._jobs.get(job_id)
            if not job or job["status"] != "running":
                return

            if progress.get("percent") is not None:
                job["progress_percent"] = progress["percent"]
            if progress.get("eta_sec") is not None:
                job["eta_sec"] = progress["eta_sec"]

            self._persist_locked()

    def _persist_locked(self) -> None:
        jobs = [dict(self._jobs[job_id]) for job_id in self._job_order if job_id in self._jobs]
        self.history_store.save(jobs)

    def _mark_canceled_locked(self, job: dict[str, Any]) -> None:
        job["status"] = "canceled"
        job["finished_at"] = _utc_now_iso()
        job["error_message"] = "已取消。"
        job["eta_sec"] = None
        self._persist_locked()

    def _terminate_process(self, process: sp.Popen[str]) -> None:
        if process.poll() is not None:
            return
        try:
            process.terminate()
            process.wait(timeout=2)
        except Exception:
            try:
                process.kill()
            except Exception:
                pass

    def _parse_progress(self, line: str) -> dict[str, Any] | None:
        percent = None
        eta_sec = None

        percent_match = _PROGRESS_RE.search(line)
        if percent_match:
            try:
                percent = float(percent_match.group("percent"))
            except (TypeError, ValueError):
                percent = None

        eta_match = _ETA_RE.search(line)
        if eta_match:
            eta_sec = self._parse_eta_to_seconds(eta_match.group("eta"))

        if percent is None and eta_sec is None:
            return None

        return {"percent": percent, "eta_sec": eta_sec}

    def _parse_output_path(self, line: str) -> str:
        for pattern in _DESTINATION_PATTERNS:
            match = pattern.search(line)
            if match:
                path = match.group("path").strip().strip('"')
                return path
        return ""

    def _to_media_relative(self, path_text: str) -> str:
        if not path_text:
            return ""
        try:
            absolute = Path(path_text).resolve()
            media_root = self.media_root.resolve()
            if str(absolute).startswith(str(media_root)):
                return str(absolute.relative_to(media_root))
        except Exception:
            return ""
        return ""

    def _parse_eta_to_seconds(self, value: str) -> int | None:
        parts = value.split(":")
        if not parts:
            return None
        try:
            nums = [int(part) for part in parts]
        except ValueError:
            return None

        if len(nums) == 2:
            return nums[0] * 60 + nums[1]
        if len(nums) == 3:
            return nums[0] * 3600 + nums[1] * 60 + nums[2]
        return None

    def _normalize_execute_result(self, result: Any) -> tuple[int, str, list[str]]:
        if not isinstance(result, tuple):
            return 1, "下载器返回结果格式错误。", []

        return_code = _to_int(result[0]) if len(result) >= 1 else 1
        if return_code is None:
            return_code = 1
        error_message = str(result[1] or "") if len(result) >= 2 else ""
        raw_output = result[2] if len(result) >= 3 else ""

        if isinstance(raw_output, str):
            outputs = [raw_output] if raw_output else []
        elif isinstance(raw_output, list):
            outputs = [str(item).strip() for item in raw_output if str(item).strip()]
        else:
            outputs = []
        return return_code, error_message, outputs

    def _probe_binary(self, command: str) -> tuple[str | None, str]:
        binary_path = shutil.which(command)
        if not binary_path:
            return None, ""
        try:
            out = sp.check_output([binary_path, "--version"], text=True, timeout=3)
            return binary_path, (out or "").strip().splitlines()[0]
        except Exception:
            return binary_path, ""

    def _extract_error_line(self, line: str) -> str:
        lowered = line.lower()
        if "error:" in lowered:
            return line
        return ""

    def _expand_user_path(self, path_text: str) -> Path:
        return Path(path_text).expanduser()

    def _collect_gallery_outputs(self, temp_dir: Path, *, excluded: set[Path]) -> list[Path]:
        outputs: list[Path] = []
        for entry in temp_dir.rglob("*"):
            if not entry.is_file():
                continue
            resolved = entry.resolve()
            if resolved in excluded:
                continue
            if resolved.name.endswith(".part"):
                continue
            outputs.append(resolved)
        outputs.sort(key=lambda p: str(p))
        return outputs

    def _move_file_to_media_root(self, source_path: Path) -> str:
        if not source_path.exists() or not source_path.is_file():
            return ""

        target_name = source_path.name
        target = (self.media_root / target_name).resolve()
        target = self._next_available_path(target)
        try:
            source_path.replace(target)
        except OSError:
            shutil.move(str(source_path), str(target))
        return self._to_media_relative(str(target))

    def _next_available_path(self, path: Path) -> Path:
        if not path.exists():
            return path
        stem = path.stem
        suffix = path.suffix
        parent = path.parent
        idx = 1
        while True:
            candidate = parent / f"{stem} ({idx}){suffix}"
            if not candidate.exists():
                return candidate
            idx += 1

    def _record_job_sources_on_success(self, context: dict[str, Any]) -> None:
        if not self.source_store:
            return

        files = context.get("files")
        if not isinstance(files, list):
            return
        raw_url = str(context.get("url") or "").strip()
        if not raw_url:
            return

        base_meta = _normalize_source_meta(
            {
                "source_url_raw": raw_url,
                "source_url_display": _strip_tracking_query(raw_url),
                "source_domain": _derive_source_domain(raw_url),
                "engine": str(context.get("engine") or "").strip(),
                "job_id": str(context.get("job_id") or "").strip(),
                "created_at": str(context.get("created_at") or _utc_now_iso()),
            },
            resolved_by="map",
        )
        if not base_meta:
            return

        records: dict[str, dict[str, Any]] = {}
        for rel in files:
            key = _normalize_file_rel(str(rel or ""))
            if not key:
                continue
            records[key] = dict(base_meta)
        if records:
            self.source_store.set_many(records)

    def _resolve_source_from_map(self, file_rel: str) -> dict[str, Any] | None:
        if not self.source_store:
            return None
        mapped = self.source_store.get(file_rel)
        if not mapped:
            return None
        return _normalize_source_meta(mapped, resolved_by="map")

    def _resolve_source_from_info_json(self, file_rel: str) -> dict[str, Any] | None:
        media_file = self._resolve_media_file_path(file_rel)
        if not media_file or not media_file.exists():
            return None

        candidates = [
            media_file.with_suffix(".info.json"),
            Path(str(media_file) + ".info.json"),
        ]
        data: dict[str, Any] | None = None
        for path in candidates:
            if not path.exists() or not path.is_file():
                continue
            try:
                with path.open("r", encoding="utf-8") as f:
                    loaded = json.load(f)
                if isinstance(loaded, dict):
                    data = loaded
                    break
            except Exception:
                continue
        if not data:
            return None

        raw_url = ""
        for field in ("webpage_url", "original_url", "url"):
            value = str(data.get(field) or "").strip()
            if value:
                raw_url = value
                break
        if not raw_url:
            return None

        meta = _normalize_source_meta(
            {
                "source_url_raw": raw_url,
                "source_url_display": _strip_tracking_query(raw_url),
                "source_domain": _derive_source_domain(raw_url),
                "engine": str(data.get("extractor_key") or "").strip(),
                "created_at": _utc_now_iso(),
            },
            resolved_by="infojson",
        )
        if meta and self.source_store:
            self.source_store.set_many({file_rel: dict(meta)})
        return meta

    def _resolve_source_from_filename(self, file_rel: str) -> dict[str, Any] | None:
        name = Path(file_rel).name
        stem = Path(name).stem

        if "__" in stem:
            parts = stem.split("__")
            if len(parts) >= 4:
                extractor = parts[0].strip().lower()
                uploader_id = parts[1].strip()
                display_id = parts[2].strip()
                media_id = parts[3].strip()
                display_id = "" if display_id.lower() == "na" else display_id
                media_id = "" if media_id.lower() == "na" else media_id
                uploader_id = "" if uploader_id.lower() == "na" else uploader_id

                if extractor in {"twitter", "x"} and display_id:
                    url = f"https://x.com/i/web/status/{display_id}"
                    if uploader_id:
                        url = f"https://x.com/{uploader_id}/status/{display_id}"
                    return _normalize_source_meta(
                        {
                            "source_url_raw": url,
                            "source_url_display": url,
                            "source_domain": "x.com",
                            "engine": "yt-dlp",
                        },
                        resolved_by="filename",
                    )

                if extractor.startswith("youtube") and media_id:
                    url = f"https://www.youtube.com/watch?v={media_id}"
                    return _normalize_source_meta(
                        {
                            "source_url_raw": url,
                            "source_url_display": url,
                            "source_domain": "youtube.com",
                            "engine": "yt-dlp",
                        },
                        resolved_by="filename",
                    )

                if extractor.startswith("tiktok") and media_id:
                    url = f"https://www.tiktok.com/@_/video/{media_id}"
                    return _normalize_source_meta(
                        {
                            "source_url_raw": url,
                            "source_url_display": url,
                            "source_domain": "tiktok.com",
                            "engine": "yt-dlp",
                        },
                        resolved_by="filename",
                    )

                if extractor.startswith("instagram") and display_id:
                    url = f"https://www.instagram.com/p/{display_id}/"
                    return _normalize_source_meta(
                        {
                            "source_url_raw": url,
                            "source_url_display": url,
                            "source_domain": "instagram.com",
                            "engine": "yt-dlp",
                        },
                        resolved_by="filename",
                    )

        old_match = _OLD_TEMPLATE_ID_RE.search(name)
        if old_match:
            old_id = old_match.group("id").strip()
            if old_id.isdigit():
                url = f"https://x.com/i/web/status/{old_id}"
                return _normalize_source_meta(
                    {
                        "source_url_raw": url,
                        "source_url_display": url,
                        "source_domain": "x.com",
                        "engine": "yt-dlp",
                    },
                    resolved_by="filename",
                )
        return None

    def _resolve_media_file_path(self, file_rel: str) -> Path | None:
        key = _normalize_file_rel(file_rel)
        if not key:
            return None
        path = (self.media_root / key).resolve()
        try:
            path.relative_to(self.media_root)
        except ValueError:
            return None
        return path

    def _resolve_cookie_choice(self, *, url: str, cookie_mode: str, cookie_file: str) -> tuple[str, str, str | None]:
        with self._lock:
            cookie_enabled = bool(self._config.get("cookie_enabled", True))
            cookie_dir_text = str(self._config.get("cookie_dir", DEFAULT_DOWNLOAD_CONFIG["cookie_dir"])).strip()

        if not cookie_enabled:
            return "", "none", None
        if cookie_mode == "none":
            return "", "none", None

        info = self.list_cookie_files()
        files = info.get("files") or []
        cookie_dir = self._expand_cookie_dir(cookie_dir_text)
        if not files:
            if cookie_mode == "manual":
                return "", "none", f"未找到 cookie 文件目录或文件: {cookie_dir}"
            return "", "none", None

        if cookie_mode == "manual":
            if not cookie_file:
                return "", "none", "手动模式缺少 cookie_file。"
            if cookie_file not in files:
                return "", "none", f"cookie 文件不存在: {cookie_file}"
            return cookie_file, "manual", None

        parsed = urlparse(url)
        host = (parsed.hostname or "").strip().lower()
        if not host:
            return "", "none", None

        lowered_files = {name.lower(): name for name in files}
        for candidate in _domain_candidates(host):
            hits = [lowered_files[key] for key in lowered_files if candidate in key]
            if hits:
                hits.sort(key=lambda item: (len(item), item.lower()))
                return hits[0], "auto", None
        return "", "none", None

    def _expand_cookie_dir(self, cookie_dir_text: str) -> Path:
        return Path(cookie_dir_text).expanduser()

    def _cookie_dir_path(self) -> Path:
        with self._lock:
            cookie_dir_text = str(self._config.get("cookie_dir", DEFAULT_DOWNLOAD_CONFIG["cookie_dir"])).strip()
        cookie_dir = self._expand_cookie_dir(cookie_dir_text)
        cookie_dir.mkdir(parents=True, exist_ok=True)
        return cookie_dir

    def _resolve_cookie_path(self, cookie_file: str) -> tuple[Path | None, str | None]:
        info = self.list_cookie_files()
        cookie_dir = Path(info["cookie_dir"]).expanduser().resolve()
        if cookie_file not in info.get("files", []):
            return None, f"cookie 文件不存在: {cookie_file}"
        path = (cookie_dir / cookie_file).resolve()
        try:
            path.relative_to(cookie_dir)
        except ValueError:
            return None, "cookie 文件路径非法。"
        if not path.exists() or not path.is_file():
            return None, f"cookie 文件不可用: {cookie_file}"
        return path, None



import argparse
import datetime
import hashlib
import json
import os
import subprocess as sp
import sys
from pathlib import Path
import mimetypes
from tiklocal.paths import get_thumbnails_dir, get_thumbs_map_path, get_data_dir


def _thumb_key(rel_path: str) -> str:
    return hashlib.sha1(rel_path.encode('utf-8', errors='ignore')).hexdigest() + '.jpg'


def _thumb_path(rel_path: str) -> Path:
    return get_thumbnails_dir() / _thumb_key(rel_path)


def _map_path() -> Path:
    return get_thumbs_map_path()


def _load_map() -> dict:
    p = _map_path()
    if p.exists():
        try:
            return json.loads(p.read_text(encoding='utf-8'))
        except Exception:
            return {}
    return {}


def _save_map(data: dict) -> None:
    p = _map_path()
    p.write_text(json.dumps(data, ensure_ascii=False), encoding='utf-8')


def _probe_duration(path: Path) -> float | None:
    # 尝试用 ffprobe 获取时长（秒）
    cmd = [
        'ffprobe', '-v', 'error', '-show_entries', 'format=duration',
        '-of', 'default=noprint_wrappers=1:nokey=1', str(path)
    ]
    try:
        out = sp.check_output(cmd, stderr=sp.DEVNULL, timeout=10)
        val = float(out.decode().strip())
        if val > 0:
            return val
    except Exception:
        return None
    return None


def _ffmpeg_capture(input_path: Path, output_path: Path, ts: float | None) -> bool:
    candidates: list[float]
    if ts is not None and ts >= 0:
        candidates = [ts]
    else:
        dur = _probe_duration(input_path)
        if dur and dur > 1:
            t = max(1.0, min(dur - 1.0, dur * 0.2))
            candidates = [t, 5.0, 1.0, 0.1]
        else:
            candidates = [5.0, 1.0, 0.1]

    for t in candidates:
        cmd = [
            'ffmpeg', '-y', '-ss', str(max(0.0, float(t))), '-i', str(input_path),
            '-frames:v', '1', '-vf', 'scale=-1:360:force_original_aspect_ratio=decrease',
            '-q:v', '3', str(output_path)
        ]
        try:
            sp.run(cmd, stdout=sp.DEVNULL, stderr=sp.DEVNULL, timeout=30)
            if output_path.exists() and output_path.stat().st_size > 0:
                return True
        except Exception:
            continue
    return False


VIDEO_EXTS = {'.mp4', '.webm', '.mov', '.mkv', '.avi', '.m4v'}


def _is_video(path: Path) -> bool:
    suf = path.suffix.lower()
    if suf in VIDEO_EXTS:
        return True
    try:
        mime = mimetypes.guess_type(path.name)[0] or ''
    except Exception:
        mime = ''
    return mime.startswith('video/')


def _iter_videos(root: Path) -> list[Path]:
    videos: list[Path] = []
    # 先按扩展名快速匹配，再做 mimetype 二次校验
    for pattern in ('*.mp4', '*.webm', '*.mov', '*.mkv', '*.avi', '*.m4v',
                    '*.MP4', '*.WEBM', '*.MOV', '*.MKV', '*.AVI', '*.M4V'):
        for p in root.glob(f'**/{pattern}'):
            if _is_video(p):
                videos.append(p)
    return videos


def _print_progress(current: int, total: int, prefix: str = '') -> None:
    width = 28
    filled = int(width * current / total) if total else width
    bar = '█' * filled + '─' * (width - filled)
    percent = (current / total * 100) if total else 100
    sys.stdout.write(f"\r{prefix}[{bar}] {current}/{total} {percent:5.1f}%")
    sys.stdout.flush()


def generate_thumbnails(media_root: str | Path, overwrite: bool = False, limit: int = 0, show_progress: bool = True) -> dict:
    root = Path(media_root)
    mapping = _load_map()
    videos = _iter_videos(root)
    total = len(videos)
    done = 0
    skipped = 0
    failed = 0
    limit_left = limit if limit and limit > 0 else None

    if show_progress:
        print(f'数据目录: {get_data_dir()}')
        print(f'发现视频 {total} 个，缩略图目录：{get_thumbnails_dir()}')

    processed = 0
    for vp in videos:
        rel = str(vp.relative_to(root))
        out = _thumb_path(rel)
        # 二次校验，防御性判断
        if not _is_video(vp):
            skipped += 1
            processed += 1
            if show_progress:
                _print_progress(processed, total, prefix='生成中 ')
            continue

        if out.exists() and not overwrite:
            skipped += 1
            processed += 1
            if show_progress:
                _print_progress(processed, total, prefix='生成中 ')
            if limit_left is not None:
                limit_left -= 1
                if limit_left <= 0:
                    break
            continue

        ok = _ffmpeg_capture(vp, out, mapping.get(rel, {}).get('ts'))
        if ok:
            mapping[rel] = {
                'ts': mapping.get(rel, {}).get('ts'),
                'updated_at': datetime.datetime.now().isoformat(timespec='seconds')
            }
            done += 1
        else:
            failed += 1

        processed += 1
        if show_progress:
            _print_progress(processed, total, prefix='生成中 ')

        if limit_left is not None:
            limit_left -= 1
            if limit_left <= 0:
                break

    _save_map(mapping)
    if show_progress:
        print()  # 换行
        print(f'完成：生成 {done}，跳过 {skipped}，失败 {failed}，总计 {total}')
    return {
        'total': total,
        'generated': done,
        'skipped': skipped,
        'failed': failed,
    }


def clean_thumbnails(media_root: str | Path, show_progress: bool = True) -> dict:
    root = Path(media_root)
    mapping = _load_map()
    keys = list(mapping.keys())
    total = len(keys)
    removed = 0
    kept = 0

    if show_progress:
        print(f'开始清理：映射 {total} 条')

    for i, rel in enumerate(keys, start=1):
        target = (root / rel)
        thumb = _thumb_path(rel)
        invalid = (not target.exists()) or (not _is_video(target))
        if invalid:
            try:
                if thumb.exists():
                    thumb.unlink()
            except Exception:
                pass
            mapping.pop(rel, None)
            removed += 1
        else:
            kept += 1
        if show_progress:
            _print_progress(i, total, prefix='清理中 ')

    _save_map(mapping)
    if show_progress:
        print()  # 换行
        print(f'清理完成：保留 {kept}，移除 {removed}，总计 {total}')
    return {'kept': kept, 'removed': removed, 'total': total}


def verify_thumbnails(media_root: str | Path) -> dict:
    root = Path(media_root)
    mapping = _load_map()
    videos = _iter_videos(root)
    video_set = {str(p.relative_to(root)) for p in videos}

    mapped = 0
    invalid = 0
    missing = 0

    for rel, meta in mapping.items():
        thumb = _thumb_path(rel)
        if rel in video_set and thumb.exists():
            mapped += 1
        else:
            invalid += 1

    for rel in video_set:
        if not _thumb_path(rel).exists():
            missing += 1

    print(f"视频总数: {len(video_set)}  | 已有缩略图: {mapped}  | 异常映射: {invalid}  | 待生成: {missing}")
    return {
        'videos': len(video_set),
        'mapped': mapped,
        'invalid': invalid,
        'missing': missing,
    }


def main():
    parser = argparse.ArgumentParser(description='TikLocal 缩略图工具')
    parser.add_argument('media_root', nargs='?', help='媒体根目录（可省略以使用环境变量 MEDIA_ROOT）')
    parser.add_argument('--overwrite', action='store_true', help='已存在时覆盖重建')
    parser.add_argument('--limit', type=int, default=0, help='最多处理多少个（0 表示全部）')
    parser.add_argument('--clean', action='store_true', help='清理非视频/孤儿缩略图与映射')
    parser.add_argument('--verify', action='store_true', help='仅校验覆盖率与异常，不修改')
    args = parser.parse_args()

    media_root = args.media_root or os.environ.get('MEDIA_ROOT')
    if not media_root:
        parser.error('必须指定媒体目录：位置参数或环境变量 MEDIA_ROOT')

    root = Path(media_root)
    if not root.exists() or not root.is_dir():
        parser.error(f'媒体目录不可用：{media_root}')

    if args.clean:
        clean_thumbnails(root, show_progress=True)
        return
    if args.verify:
        verify_thumbnails(root)
        return
    generate_thumbnails(root, overwrite=args.overwrite, limit=args.limit, show_progress=True)


if __name__ == '__main__':
    main()



import datetime
import json
import random
import subprocess as sp
from pathlib import Path
from typing import Callable
from urllib.parse import quote

from PIL import Image


def build_feed_media_item(name: str, media_type: str) -> dict[str, str]:
    encoded = quote(name)
    detail_url = f"/detail/{encoded}" if media_type == 'video' else f"/image?uri={encoded}"
    media_url = f"/media?uri={encoded}"
    return {
        'type': media_type,
        'name': name,
        'media_url': media_url,
        'thumb_url': f"/thumb?uri={encoded}" if media_type == 'video' else media_url,
        'detail_url': detail_url,
    }


def collect_source_media_groups(
    records: list[dict],
    download_source_store,
    build_item: Callable[[str, str], dict[str, str]] = build_feed_media_item,
) -> list[dict]:
    records_by_name = {str(item.get('name') or ''): item for item in records if item.get('name')}
    source_map = download_source_store.get_many(list(records_by_name.keys()))
    groups_by_source: dict[str, dict] = {}
    groups_by_job: dict[str, dict] = {}

    for name, source_meta in source_map.items():
        if not isinstance(source_meta, dict):
            continue
        record = records_by_name.get(name)
        if not record:
            continue

        normalized_item = {
            'name': name,
            'media_type': str(record.get('media_type') or ''),
            'sort_ts': float(record.get('mtime_ts') or 0),
        }

        source_url = str(source_meta.get('source_url_display') or source_meta.get('source_url_raw') or '').strip()
        if source_url:
            group = groups_by_source.setdefault(source_url, {
                'key': source_url,
                'source_domain': str(source_meta.get('source_domain') or '').strip(),
                'created_at': str(source_meta.get('created_at') or '').strip(),
                'items': [],
            })
            group['items'].append(normalized_item)
            if str(source_meta.get('created_at') or '').strip() > str(group.get('created_at') or ''):
                group['created_at'] = str(source_meta.get('created_at') or '').strip()

        job_id = str(source_meta.get('job_id') or '').strip()
        if job_id:
            group = groups_by_job.setdefault(job_id, {
                'key': job_id,
                'source_domain': str(source_meta.get('source_domain') or '').strip(),
                'created_at': str(source_meta.get('created_at') or '').strip(),
                'items': [],
            })
            group['items'].append(normalized_item)
            if str(source_meta.get('created_at') or '').strip() > str(group.get('created_at') or ''):
                group['created_at'] = str(source_meta.get('created_at') or '').strip()

    results: list[dict] = []
    seen_group_signatures: set[tuple[str, ...]] = set()
    for group in list(groups_by_source.values()) + list(groups_by_job.values()):
        entries = group.get('items') or []
        unique_names = sorted({str(item.get('name') or '') for item in entries if item.get('name')})
        if len(unique_names) < 2:
            continue
        signature = tuple(unique_names)
        if signature in seen_group_signatures:
            continue
        seen_group_signatures.add(signature)

        sorted_items = sorted(
            [item for item in entries if item.get('name') in records_by_name],
            key=lambda item: (float(item.get('sort_ts') or 0), str(item.get('name') or '')),
            reverse=True,
        )
        domain = str(group.get('source_domain') or '').strip()
        results.append({
            'source_domain': domain,
            'created_at': str(group.get('created_at') or ''),
            'items': [
                build_item(str(item['name']), str(item['media_type']))
                for item in sorted_items[:8]
            ],
        })

    results.sort(
        key=lambda item: (
            str(item.get('created_at') or ''),
            len(item.get('items') or []),
        ),
        reverse=True,
    )
    return results


def collect_library_records(
    library_service,
    favorite_service,
    image_extensions: set[str],
    *,
    favorites_only: bool = False,
) -> list[dict]:
    favorite_set = favorite_service.load()
    all_paths = library_service.scan_videos() + library_service.scan_images()

    seen_rel: set[str] = set()
    records_by_identity: dict[tuple, dict] = {}

    for path in all_paths:
        try:
            rel = library_service.get_relative_path(path)
            if rel in seen_rel:
                continue
            seen_rel.add(rel)

            stat = path.stat()
            identity = (int(getattr(stat, 'st_dev', 0)), int(getattr(stat, 'st_ino', 0)))
            if identity == (0, 0):
                identity = ('resolved', str(path.resolve()))

            media_type = 'image' if path.suffix.lower() in image_extensions else 'video'
            candidate = {
                'name': rel,
                'media_type': media_type,
                'mtime_ts': float(stat.st_mtime),
                'size_bytes': int(stat.st_size),
                'is_favorite': rel in favorite_set,
            }

            existing = records_by_identity.get(identity)
            if existing is None:
                records_by_identity[identity] = candidate
                continue

            if candidate['is_favorite'] and not existing['is_favorite']:
                records_by_identity[identity] = candidate
                continue

            if (
                candidate['mtime_ts'] > existing['mtime_ts']
                or (
                    candidate['mtime_ts'] == existing['mtime_ts']
                    and candidate['name'] < existing['name']
                )
            ):
                records_by_identity[identity] = candidate
        except Exception:
            continue

    records = list(records_by_identity.values())
    if favorites_only:
        records = [item for item in records if item.get('is_favorite')]

    records.sort(key=lambda item: (item['mtime_ts'], item['name']), reverse=True)
    return records


def build_theme_strip_candidates(
    records: list[dict],
    download_history_store,
    build_item: Callable[[str, str], dict[str, str]] = build_feed_media_item,
) -> list[dict]:
    records_by_name = {str(item.get('name') or ''): item for item in records if item.get('name')}
    candidates: list[dict] = []

    favorite_records = [
        item for item in records
        if item.get('is_favorite') and item.get('media_type') in {'video', 'image'}
    ]
    favorite_records.sort(key=lambda item: item.get('mtime_ts') or 0, reverse=True)
    favorite_items = [
        build_item(item['name'], item['media_type'])
        for item in favorite_records[:8]
    ]
    if len(favorite_items) >= 3:
        candidates.append({
            'type': 'theme_strip',
            'name': 'theme:favorite-picks',
            'title': '收藏精选',
            'subtitle': '快速跳去收藏页继续看。',
            'target_url': '/favorite',
            'target_label': '打开收藏',
            'items': favorite_items,
        })

    history = download_history_store.get()
    recent_download_names: list[str] = []
    seen_names: set[str] = set()
    sorted_history = sorted(
        [item for item in history if isinstance(item, dict)],
        key=lambda item: str(item.get('created_at') or ''),
        reverse=True,
    )
    for job in sorted_history:
        if str(job.get('status') or '') != 'success':
            continue
        raw_files = job.get('output_files_rel')
        if not isinstance(raw_files, list):
            continue
        for value in raw_files:
            name = str(value or '').strip().replace('\\', '/')
            while name.startswith('./'):
                name = name[2:]
            if not name or name in seen_names or name not in records_by_name:
                continue
            seen_names.add(name)
            recent_download_names.append(name)
            if len(recent_download_names) >= 8:
                break
        if len(recent_download_names) >= 8:
            break

    download_items = [
        build_item(name, str(records_by_name[name]['media_type']))
        for name in recent_download_names
    ]
    if len(download_items) >= 3:
        candidates.append({
            'type': 'theme_strip',
            'name': 'theme:recent-downloads',
            'title': '最近下载',
            'subtitle': '快速跳去媒体库继续看。',
            'target_url': '/library',
            'target_label': '打开媒体库',
            'items': download_items,
        })

    return candidates


def read_media_dims_from_metadata(metadata_store, name: str) -> tuple[int | None, int | None]:
    payload = metadata_store.get(name)
    if not isinstance(payload, dict):
        return None, None
    media_meta = payload.get('media_meta')
    if not isinstance(media_meta, dict):
        return None, None
    try:
        width = int(media_meta.get('width') or 0)
        height = int(media_meta.get('height') or 0)
    except (TypeError, ValueError):
        return None, None
    if width <= 0 or height <= 0:
        return None, None
    return width, height


def save_media_dims_to_metadata(metadata_store, name: str, media_type: str, width: int, height: int) -> None:
    if width <= 0 or height <= 0:
        return
    current = metadata_store.get(name)
    payload = dict(current) if isinstance(current, dict) else {}
    payload['media_meta'] = {
        'type': media_type,
        'width': int(width),
        'height': int(height),
        'updated_at': datetime.datetime.utcnow().isoformat() + 'Z',
    }
    metadata_store.set(name, payload, overwrite=True)


def probe_media_dims(library_service, name: str, media_type: str) -> tuple[int | None, int | None]:
    target = library_service.resolve_path(name)
    if not target or not target.exists():
        return None, None

    if media_type == 'image':
        try:
            with Image.open(target) as img:
                width, height = img.size
            if int(width) > 0 and int(height) > 0:
                return int(width), int(height)
        except Exception:
            return None, None
        return None, None

    try:
        cmd = [
            'ffprobe',
            '-v', 'error',
            '-select_streams', 'v:0',
            '-show_entries', 'stream=width,height',
            '-of', 'json',
            str(target),
        ]
        proc = sp.run(cmd, capture_output=True, text=True, timeout=8)
        if proc.returncode != 0:
            return None, None
        payload = json.loads(proc.stdout or '{}')
        streams = payload.get('streams') or []
        if not streams:
            return None, None
        stream = streams[0] if isinstance(streams[0], dict) else {}
        width = int(stream.get('width') or 0)
        height = int(stream.get('height') or 0)
        if width > 0 and height > 0:
            return width, height
    except Exception:
        return None, None
    return None, None


def get_or_probe_media_dims(metadata_store, library_service, name: str, media_type: str) -> tuple[int | None, int | None]:
    width, height = read_media_dims_from_metadata(metadata_store, name)
    if width and height:
        return width, height
    width, height = probe_media_dims(library_service, name, media_type)
    if width and height:
        save_media_dims_to_metadata(metadata_store, name, media_type, width, height)
    return width, height


def serialize_library_item(record: dict, metadata_store, library_service) -> dict:
    name = str(record.get('name') or '')
    media_type = str(record.get('media_type') or 'video')
    width, height = get_or_probe_media_dims(metadata_store, library_service, name, media_type)
    encoded = quote(name)
    media_url = f"/media?uri={encoded}"
    return {
        'name': name,
        'type': media_type,
        'media_url': media_url,
        'detail_url': f"/image?uri={encoded}" if media_type == 'image' else f"/detail/{encoded}",
        'thumb_url': media_url if media_type == 'image' else f"/thumb?uri={encoded}",
        'mtime_ts': float(record.get('mtime_ts') or 0),
        'size_bytes': int(record.get('size_bytes') or 0),
        'width': int(width) if width else None,
        'height': int(height) if height else None,
    }


def collect_collection_records(
    collection_id: str,
    collection_store,
    favorite_service,
    library_service,
    image_extensions: set[str],
) -> tuple[dict | None, list[dict]]:
    collection = collection_store.get(collection_id)
    if not collection:
        return None, []

    favorites = favorite_service.load()
    uris = collection_store.list_item_uris(collection_id, newest_first=True)
    records: list[dict] = []
    seen: set[str] = set()
    for uri in uris:
        if uri in seen:
            continue
        seen.add(uri)
        target = library_service.resolve_path(uri)
        if not target or not target.exists():
            continue
        try:
            stat = target.stat()
        except OSError:
            continue
        media_type = 'image' if target.suffix.lower() in image_extensions else 'video'
        records.append({
            'name': uri,
            'media_type': media_type,
            'mtime_ts': float(stat.st_mtime),
            'size_bytes': int(stat.st_size),
            'is_favorite': uri in favorites,
        })
    return collection, records


def collection_cover_payload(collection: dict, collection_store, library_service, image_extensions: set[str]) -> tuple[str, str]:
    cover_uri = str(collection.get('cover_uri') or '').strip()
    if cover_uri:
        target = library_service.resolve_path(cover_uri)
        if target and target.exists():
            media_type = 'image' if target.suffix.lower() in image_extensions else 'video'
            return cover_uri, media_type

    uris = collection_store.list_item_uris(str(collection.get('id') or ''), newest_first=True)
    for uri in uris:
        target = library_service.resolve_path(uri)
        if not target or not target.exists():
            continue
        media_type = 'image' if target.suffix.lower() in image_extensions else 'video'
        return uri, media_type
    return '', ''


def serialize_collection_summary(collection: dict, collection_store, library_service, image_extensions: set[str]) -> dict:
    collection_id = str(collection.get('id') or '')
    cover_uri, cover_type = collection_cover_payload(collection, collection_store, library_service, image_extensions)
    cover_encoded = quote(cover_uri, safe='') if cover_uri else ''
    cover_media_url = f"/media?uri={cover_encoded}" if cover_uri else ''
    return {
        'id': collection_id,
        'name': str(collection.get('name') or ''),
        'description': str(collection.get('description') or ''),
        'item_count': int(collection.get('item_count') or 0),
        'cover_uri': cover_uri,
        'cover_type': cover_type,
        'cover_media_url': cover_media_url,
        'detail_url': f"/collection/{quote(collection_id, safe='')}" if collection_id else '#',
        'created_at': str(collection.get('created_at') or ''),
        'updated_at': str(collection.get('updated_at') or ''),
    }


def apply_library_mode(records: list[dict], *, mode: str, min_mb: int, seed: str) -> list[dict]:
    if mode == 'image_random':
        image_records = [item for item in records if item.get('media_type') == 'image']
        rng = random.Random(seed)
        rng.shuffle(image_records)
        return image_records
    if mode == 'video_latest':
        return [item for item in records if item.get('media_type') == 'video']
    if mode == 'big_files':
        min_bytes = max(1, int(min_mb)) * 1024 * 1024
        video_records = [item for item in records if item.get('media_type') == 'video' and int(item.get('size_bytes') or 0) >= min_bytes]
        video_records.sort(key=lambda item: (item['size_bytes'], item['mtime_ts'], item['name']), reverse=True)
        return video_records
    return records


def build_library_page(
    *,
    favorites_only: bool = False,
    mode: str = 'all',
    offset: int = 0,
    limit: int = 48,
    min_mb: int = 50,
    seed: str = '',
    collection_id: str = '',
    collect_collection_records_fn: Callable[[str], tuple[dict | None, list[dict]]],
    collect_library_records_fn: Callable[..., list[dict]],
    apply_library_mode_fn: Callable[..., list[dict]] = apply_library_mode,
    serialize_library_item_fn: Callable[[dict], dict],
) -> dict:
    records: list[dict] = []
    if collection_id:
        _, records = collect_collection_records_fn(collection_id)
    else:
        records = collect_library_records_fn(favorites_only=favorites_only)
        records = apply_library_mode_fn(records, mode=mode, min_mb=min_mb, seed=seed)
    total = len(records)
    start = max(0, int(offset))
    safe_limit = max(12, min(int(limit), 96))
    end = start + safe_limit
    items = [serialize_library_item_fn(record) for record in records[start:end]]
    return {
        'items': items,
        'total': total,
        'offset': start,
        'limit': safe_limit,
        'next_offset': end,
        'has_more': end < total,
        'seed': seed,
    }


def build_mix_feed_page(
    *,
    page: int,
    size: int,
    seed: str,
    recommend_service,
    collect_library_records_fn: Callable[..., list[dict]],
    build_theme_strip_candidates_fn: Callable[[list[dict]], list[dict]],
    collect_source_media_groups_fn: Callable[[list[dict]], list[dict]],
    build_feed_media_item_fn: Callable[[str, str], dict[str, str]] = build_feed_media_item,
) -> dict:
    video_ratio = 4
    image_ratio = 1

    ratio_total = video_ratio + image_ratio
    end = page * size
    start = max(0, end - size)
    request_window = end + size
    video_need = max(size, int(request_window * (video_ratio / ratio_total)) + video_ratio * 2)
    image_need = max(size, int(request_window * (image_ratio / ratio_total)) + image_ratio * 2)

    videos = recommend_service.get_weighted_selection(
        file_type='video',
        limit=video_need,
        seed=f"{seed}:video",
    )
    images = recommend_service.get_weighted_selection(
        file_type='image',
        limit=image_need,
        seed=f"{seed}:image",
    )

    target_image_prob = image_ratio / max(1, ratio_total)
    max_video_streak = 6
    max_image_streak = 2
    rng = random.Random(f"{seed}:mix")

    mixed: list[tuple[str, str]] = []
    seen: set[str] = set()
    vi = 0
    ii = 0
    used_video = 0
    used_image = 0
    video_streak = 0
    image_streak = 0

    def pick_next_available(kind: str) -> tuple[str, str]:
        nonlocal vi, ii
        if kind == 'video':
            while vi < len(videos):
                name = videos[vi]
                vi += 1
                if name:
                    return 'video', name
            return '', ''
        while ii < len(images):
            name = images[ii]
            ii += 1
            if name:
                return 'image', name
        return '', ''

    while len(mixed) < request_window and (vi < len(videos) or ii < len(images)):
        if video_streak >= max_video_streak and ii < len(images):
            want_type = 'image'
        elif image_streak >= max_image_streak and vi < len(videos):
            want_type = 'video'
        else:
            total_used = used_video + used_image
            current_image_ratio = (used_image / total_used) if total_used > 0 else target_image_prob
            correction = (target_image_prob - current_image_ratio) * 0.65
            p_image = max(0.05, min(0.5, target_image_prob + correction))
            want_type = 'image' if rng.random() < p_image else 'video'

        chosen_type, chosen_name = pick_next_available(want_type)
        if not chosen_name:
            fallback = 'video' if want_type == 'image' else 'image'
            chosen_type, chosen_name = pick_next_available(fallback)

        if not chosen_name or chosen_name in seen:
            continue
        seen.add(chosen_name)
        mixed.append((chosen_type, chosen_name))
        if chosen_type == 'video':
            used_video += 1
            video_streak += 1
            image_streak = 0
        else:
            used_image += 1
            image_streak += 1
            video_streak = 0

    records = collect_library_records_fn(favorites_only=False)
    theme_candidates = build_theme_strip_candidates_fn(records)
    source_groups = collect_source_media_groups_fn(records)
    image_group_candidate = None
    image_group_names: set[str] = set()
    for group in source_groups:
        group_items = [item for item in (group.get('items') or []) if item.get('type') == 'image']
        if len(group_items) < 2:
            continue
        image_group_candidate = {
            'type': 'image_group',
            'name': f"group:{group_items[0]['name']}",
            'title': '原始图集',
            'subtitle': '左右切换查看同一帖子里的图片。',
            'items': group_items,
        }
        image_group_names = {str(item.get('name') or '') for item in group_items}
        break

    mixed_entries: list[dict] = [
        {'type': media_type, 'name': name}
        for media_type, name in mixed
    ]
    if image_group_candidate and mixed_entries:
        mixed_entries = [
            entry for entry in mixed_entries
            if not (entry.get('type') == 'image' and str(entry.get('name') or '') in image_group_names)
        ]
        group_rng = random.Random(f"{seed}:image-group")
        insert_floor = min(2, len(mixed_entries))
        insert_ceil = min(max(insert_floor, 6), len(mixed_entries))
        insert_at = insert_floor if insert_ceil <= insert_floor else group_rng.randint(insert_floor, insert_ceil)
        mixed_entries.insert(insert_at, image_group_candidate)
    if page == 1 and theme_candidates and mixed_entries:
        theme_rng = random.Random(f"{seed}:theme-strip")
        candidate = theme_rng.choice(theme_candidates)
        insert_floor = min(6, len(mixed_entries))
        insert_ceil = min(max(insert_floor, 10), len(mixed_entries))
        insert_at = insert_floor if insert_ceil <= insert_floor else theme_rng.randint(insert_floor, insert_ceil)
        mixed_entries.insert(insert_at, candidate)

    page_items = mixed_entries[start:end]
    items = []
    for entry in page_items:
        item_type = str(entry.get('type') or '')
        if item_type == 'theme_strip':
            target_url = str(entry.get('target_url') or '').strip()
            items.append({
                'type': 'theme_strip',
                'name': str(entry.get('name') or 'theme:strip'),
                'title': str(entry.get('title') or '').strip(),
                'subtitle': str(entry.get('subtitle') or '').strip(),
                'target_url': target_url,
                'target_label': str(entry.get('target_label') or '').strip(),
                'items': [
                    {
                        'type': child.get('type'),
                        'name': child.get('name'),
                        'media_url': child.get('media_url'),
                        'thumb_url': child.get('thumb_url'),
                        'detail_url': child.get('detail_url'),
                        'focus_url': f"{target_url}?focus={quote(str(child.get('name') or ''), safe='')}" if target_url and child.get('name') else target_url,
                    }
                    for child in (entry.get('items') or [])
                    if isinstance(child, dict) and child.get('name')
                ],
            })
            continue
        if item_type == 'image_group':
            items.append({
                'type': 'image_group',
                'name': str(entry.get('name') or 'group:image'),
                'title': str(entry.get('title') or '').strip(),
                'subtitle': str(entry.get('subtitle') or '').strip(),
                'items': [
                    {
                        'type': child.get('type'),
                        'name': child.get('name'),
                        'media_url': child.get('media_url'),
                        'thumb_url': child.get('thumb_url'),
                        'detail_url': child.get('detail_url'),
                    }
                    for child in (entry.get('items') or [])
                    if isinstance(child, dict) and child.get('name')
                ],
            })
            continue

        name = str(entry.get('name') or '')
        if not name or item_type not in {'video', 'image'}:
            continue
        items.append(build_feed_media_item_fn(name, item_type))

    return {
        'items': items,
        'page': page,
        'has_more': len(mixed_entries) > end,
        'seed': seed,
    }


def normalize_collection_mutation_uris(raw: object) -> list[str]:
    if not isinstance(raw, list):
        return []
    normalized: list[str] = []
    seen: set[str] = set()
    for item in raw:
        value = str(item or '').strip().replace('\\', '/')
        while value.startswith('./'):
            value = value[2:]
        if not value or value in seen:
            continue
        seen.add(value)
        normalized.append(value)
        if len(normalized) >= 200:
            break
    return normalized


def build_library_template_context(
    *,
    menu: str,
    scope: str,
    collection_id: str,
    collection_name: str,
    active_mode: str,
    min_mb: int,
    empty_message: str,
    initial_page: dict,
) -> dict:
    return {
        'menu': menu,
        'scope': scope,
        'collection_id': collection_id,
        'collection_name': collection_name,
        'active_mode': active_mode,
        'mode_seed': initial_page['seed'],
        'min_mb': min_mb,
        'empty_message': empty_message,
        'initial_items': initial_page['items'],
        'initial_has_more': initial_page['has_more'],
        'initial_offset': initial_page['offset'],
        'initial_next_offset': initial_page['next_offset'],
        'page_size': initial_page['limit'],
    }



import os
import mimetypes
import json
import math
import random
import datetime
from pathlib import Path

VIDEO_EXTENSIONS = {'.mp4', '.webm', '.mov', '.mkv', '.avi', '.m4v'}
IMAGE_EXTENSIONS = {'.jpg', '.jpeg', '.png', '.gif', '.webp', '.bmp'}
AUDIO_EXTENSIONS = {'.mp3', '.flac', '.aac', '.m4a', '.ogg', '.opus', '.wav'}
FAVORITE_FILENAME = 'favorite.json'

class LibraryService:
    def __init__(self, media_root: str):
        self.media_root = Path(media_root)

    def _is_video(self, path: Path) -> bool:
        if path.suffix.lower() in VIDEO_EXTENSIONS:
            return True
        try:
            mime = mimetypes.guess_type(path.name)[0]
            return mime and mime.startswith('video/')
        except:
            return False

    def _is_image(self, path: Path) -> bool:
        if path.suffix.lower() in IMAGE_EXTENSIONS:
            return True
        try:
            mime = mimetypes.guess_type(path.name)[0]
            return mime and mime.startswith('image/')
        except:
            return False

    def scan_videos(self, recursive=True) -> list[Path]:
        """Scan for video files."""
        # Using glob for better performance than os.scandir recursive manual loop
        pattern = '**/*' if recursive else '*'
        videos = []
        try:
            for ext in VIDEO_EXTENSIONS:
                videos.extend(self.media_root.glob(f"{pattern}{ext}"))
            # Sort by modification time (newest first) by default
            return sorted(videos, key=lambda p: p.stat().st_mtime, reverse=True)
        except Exception:
            return []

    def scan_audios(self, recursive=True) -> list[Path]:
        """Scan for audio files."""
        pattern = '**/*' if recursive else '*'
        audios = []
        try:
            for ext in AUDIO_EXTENSIONS:
                audios.extend(self.media_root.glob(f"{pattern}{ext}"))
            return sorted(audios, key=lambda p: p.stat().st_mtime, reverse=True)
        except Exception:
            return []

    def scan_images(self, recursive=True) -> list[Path]:
        """Scan for image files."""
        pattern = '**/*' if recursive else '*'
        images = []
        try:
            for ext in IMAGE_EXTENSIONS:
                images.extend(self.media_root.glob(f"{pattern}{ext}"))
                images.extend(self.media_root.glob(f"{pattern}{ext.upper()}"))
            return images
        except Exception:
            return []
            
    def get_relative_path(self, path: Path) -> str:
        try:
            return str(path.relative_to(self.media_root))
        except ValueError:
            return str(path)

    def resolve_path(self, relative_path: str) -> Path | None:
        """Securely resolve a relative path to an absolute path within media root."""
        try:
            # Unquote in case it's URL encoded, though usually handled by Flask
            target = (self.media_root / relative_path).resolve()
            if not str(target).startswith(str(self.media_root.resolve())):
                return None
            return target
        except Exception:
            return None


class FavoriteService:
    def __init__(self, media_root: str):
        self.db_path = Path(media_root) / FAVORITE_FILENAME

    def load(self) -> set[str]:
        if not self.db_path.exists():
            return set()
        try:
            with self.db_path.open('r', encoding='utf-8') as f:
                data = json.load(f)
                return set(data) if isinstance(data, list) else set()
        except:
            return set()

    def save(self, favorites: set[str]):
        try:
            with self.db_path.open('w', encoding='utf-8') as f:
                json.dump(list(favorites), f)
        except Exception as e:
            print(f"Failed to save favorites: {e}")

    def toggle(self, filename: str) -> bool:
        """Toggle favorite status, returns new state (True=fav, False=unfav)."""
        favs = self.load()
        if filename in favs:
            favs.remove(filename)
            is_fav = False
        else:
            favs.add(filename)
            is_fav = True
        self.save(favs)
        return is_fav

    def is_favorite(self, filename: str) -> bool:
        return filename in self.load()


class RecommendService:
    def __init__(self, library_service: LibraryService, favorite_service: FavoriteService):
        self.library = library_service
        self.favorites = favorite_service

    def get_weighted_selection(self, file_type='video', limit=20, seed=None) -> list[str]:
        """Get intelligent random selection of files."""
        # 1. Get Candidates
        if file_type == 'video':
            files = self.library.scan_videos()
        else:
            files = self.library.scan_images()
        
        if not files:
            return []

        # 2. Calculate Weights
        favs = self.favorites.load()
        now = datetime.datetime.now()
        weighted_pool = []
        root = self.library.media_root

        for f in files:
            try:
                rel_path = str(f.relative_to(root))
                mtime = f.stat().st_mtime
                
                # Weight Algo: Favorites * Recency
                is_fav = rel_path in favs
                base_score = 3.0 if is_fav else 1.0
                
                age_days = max((now - datetime.datetime.fromtimestamp(mtime)).total_seconds() / 86400, 0.0)
                # Decay: newer files have higher probability
                time_score = math.exp(-age_days / 90.0) 
                
                weight = base_score * (0.1 + time_score)
                weighted_pool.append((rel_path, weight))
            except Exception:
                continue

        # 3. Weighted Random Sample
        rng = random.Random(seed) if seed else random
        result = []
        
        # Optimization: If requesting all or more than available, just shuffle and return
        if limit >= len(weighted_pool):
            result = [p[0] for p in weighted_pool]
            rng.shuffle(result)
            return result

        # Weighted selection logic
        pool = weighted_pool[:]
        while pool and len(result) < limit:
            total_weight = sum(w for _, w in pool)
            if total_weight <= 0:
                break
            
            pick = rng.random() * total_weight
            current = 0
            for i, (path, weight) in enumerate(pool):
                current += weight
                if current >= pick:
                    result.append(path)
                    pool.pop(i)
                    break
        
        return result



import os
from pathlib import Path


def get_data_dir() -> Path:
    """Return the global data directory for TikLocal.
    Default to ~/.tiklocal unless TIKLOCAL_INSTANCE is set.
    """
    base = os.environ.get('TIKLOCAL_INSTANCE')
    if base:
        p = Path(base).expanduser()
    else:
        p = Path.home() / '.tiklocal'
    p.mkdir(parents=True, exist_ok=True)
    return p


def get_thumbnails_dir() -> Path:
    d = get_data_dir() / 'thumbnails'
    d.mkdir(parents=True, exist_ok=True)
    return d


def get_thumbs_map_path() -> Path:
    return get_data_dir() / 'thumbs.json'


def get_metadata_path() -> Path:
    return get_data_dir() / 'metadata.json'


def get_prompt_config_path() -> Path:
    return get_data_dir() / 'prompt_config.json'


def get_llm_config_path() -> Path:
    return get_data_dir() / 'llm_config.json'


def get_download_config_path() -> Path:
    return get_data_dir() / 'download_config.json'


def get_download_jobs_path() -> Path:
    return get_data_dir() / 'download_jobs.json'


def get_download_sources_path() -> Path:
    return get_data_dir() / 'download_sources.json'


def get_collections_path() -> Path:
    return get_data_dir() / 'collections.json'
